# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.15.0
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2022 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

# + [markdown] slideshow={"slide_type": "slide"}
# # Code example: Wiener filter

# + [markdown] slideshow={"slide_type": "subslide"}
# ## Introduction to Information Field Theory(IFT)
# We start with the measurement equation
# $$d_i = (Rs)_i+n_i$$
# Here, $s$ is a continuous field, $d$ a discrete data vector, $n$ is the discrete noise on each data point and $R$ is the instrument response.
# In most cases, $R$ is not analytically invertible.
# IFT aims at **inverting** the above uninvertible problem in the **best possible way** using Bayesian statistics.
#
# NIFTy (Numerical Information Field Theory) is a Python framework in which IFT problems can be tackled easily.
#
# Its main interfaces are:
#
# - **Spaces**: Cartesian, 2-Spheres (Healpix, Gauss-Legendre) and their respective harmonic spaces.
# - **Fields**: Defined on spaces.
# - **Operators**: Acting on fields.
#

# + [markdown] slideshow={"slide_type": "subslide"}
# ## Wiener filter on 1D- fields in IFT
#
# ### Assumptions
# - We consider a linear response R in the measurement equation $d=Rs+n$.
# - We assume the **signal** and the **noise** prior to be **Gaussian**  $\mathcal P (s) = \mathcal G (s,S)$, $\mathcal P (n) = \mathcal G (n,N)$
# - Here $S, N$ are signal and noise covariances. Therefore they are positive definite matrices.
#
# ### Wiener filter solution
# - Making use of Bayes' theorem, the posterior is proportional to the joint probability and is given by:
#
# $$\mathcal P (s|d) \propto P(s,d) = \mathcal G(d-Rs,N) \,\mathcal G(s,S) \propto \mathcal G (s-m,D)$$
#
# - Here, $m$ is the posterior mean , $D$ is the information propagator and are defined as follows:
# $$m = Dj, \quad D = (S^{-1} +R^\dagger N^{-1} R)^{-1} $$
# - There, $j$ is the information source defined as $$ j = R^\dagger N^{-1} d.$$
#
# Let us implement this in **NIFTy!** So let's import all the packages we need.
# -

# %matplotlib inline
import numpy as np
import nifty.cl as ift
import matplotlib.pyplot as plt
plt.rcParams['figure.dpi'] = 100


# + [markdown] slideshow={"slide_type": "subslide"}
# ### Implementation in NIFTy
#
# We assume statistical **homogeneity** and **isotropy**, so the signal covariance $S$ is **translation invariant** and only depends on the **absolute value** of the distance.
# According to Wiener-Khinchin theorem, the signal covariance $S$ is diagonal in harmonic space, $$S_{kk^{\prime}} = 2 \pi \delta(k-k^{\prime}) P(k)= \text{diag}(S) \equiv \widehat{S_k}$$ and is described by a one-dimensional power spectrum.
# We assume the power spectrum to follow a power-law, $$P(k) = P_0\,\left(1+\left(\frac{k}{k_0}\right)^2\right)^{-\gamma /2},$$ with $P_0 = 2 \cdot 10^4, \ k_0 = 5, \ \gamma = 4$, thus the reconstruction starts in harmonic space.

# + slideshow={"slide_type": "-"}
def pow_spec(k):
    P0, k0, gamma = [2e4, 5, 4]
    return P0 / ((1. + (k/k0)**2)**(gamma / 2))


# -

# ### Spaces and harmonic transformations
# - We define our non-harmonic signal space to be Cartesian with $N_{pix} = 512$ being the number of grid cells.
# - To connect harmonic and non-harmonic spaces we introduce the Hartley transform $H$ that is closely related to the Fourier transform but maps $\mathbb{R}\rightarrow\mathbb{R}$.
# - The covariance S in non-harmonic space is given by $$S = H^{\dagger}\widehat{S_k} H \ .$$

# Signal space is a regular Cartesian grid space
N_pix = 512
s_space = ift.RGSpace(N_pix)

# k_space is the harmonic conjugate space of s_space
HT = ift.HartleyOperator(s_space)
k_space = HT.target

S_k = ift.create_power_operator(k_space, power_spectrum=pow_spec, sampling_dtype=float)

# Sandwich Operator implements S = HT.adjoint @ S_k @ HT and enables NIFTy to sample from S
S = ift.SandwichOperator.make(bun=HT, cheese=S_k)

# ### Synthetic Data
# - In order to demonstrate the Wiener filter, we use **synthetic data**. Therefore, we draw a sample $\tilde{s}$ from $S$. (see Sampling)
# - For simplicity we define the response operator as a unit matrix, $R = \mathbb{1}$.
# - We assume the noise covariance to be uncorrelated and constant, $N = 0.2 \cdot \mathbb{1}$ and draw a sample $\tilde{n}$.
# - Thus the synthetic data $d = R(\tilde{s}) + \tilde{n}$.

# ### Sampling
#
# - Assuming we have a distribution $\mathcal{G}(b,B)$ we can sample from and we want to draw a sample from a distritbution $\mathcal{G}(c,C)$ with covariance $C$. The two distributions are connected via the relation $C = ABA^{\dagger}.$ One can show that $c= Ab$ with $b \curvearrowleft \mathcal{G}(b,B)$	has a probability distribution with covariance $C$ as desired.
# $$ \langle cc^\dagger\rangle_{\mathcal{G}(c,C)} = \langle Ab(Ab)^\dagger\rangle_{\mathcal{G}(b,B)} = \langle Abb^\dagger A^\dagger \rangle =  A \langle bb^\dagger  \rangle A^\dagger = ABA^\dagger = C$$
# - This is also true for the case that $B = \mathbb{1}$, meaning that $\mathcal{G}(b,\mathbb{1})$ Thus $C = AA^{\dagger}$ .
# - Note that, if $C$ is diagonal, $A$ is diagonal as well.
# - It can be shown that if $C = A + B$, then $c = a + b$ with $b \curvearrowleft \mathcal{G}(b,B)$ and $a \curvearrowleft \mathcal{G}(a,A)$ has a probability distribution with covariance $C$ as desired.
# - If we can draw samples from $\mathcal{G}(a,A)$, and we want to draw a sample with the covariance $A^{-1}$, one can simply show that $c = A^{-1}a$ has a  a probability distribution with covariance $A^{-1}$.
# $$\langle c c^{\dagger} \rangle = \langle A^{-1}aa^{\dagger}(A^{-1})^{\dagger} \rangle =  A^{-1}\langle aa^{\dagger}\rangle(A^{-1})^{\dagger} = A^{-1} A(A^{-1})^{\dagger} =A^{-1}$$
# as we assume $A^{-1}$ to be Hermitian.
#
# By this brief introduction to sampling, we apply it in order to get the synthetic data.
# All of these sampling rules are implemented in NIFTy so we do not need to take care of them.

# Draw a sample from signal with covariance S.
s = S.draw_sample()

# Define the responce operator that removes the geometry meaning it removes distances and volumes.
R = ift.GeometryRemover(s_space)
# Define the data space that has an unstructured domain.
d_space = R.target

# +
noiseless_data = R(s)

# This is the multiplicative factor going from a sample with unit covariance to N.
noise_amplitude = np.sqrt(0.2)
# Define the noise covariance
N = ift.ScalingOperator(d_space, noise_amplitude**2, float)
# Draw a sample from noise with covariance N.
n = N.draw_sample()

# Synthetic data
d = noiseless_data + n
# -

# ### Information source and information propagator
#
# Now that we have the synthetic data, we are one step closer to the Wiener filter!
# In order to apply Wiener filter on the data we first need to define the information source $j$ along with the information propagator $D$.

# +
# Define the information propagator.
j = R.adjoint(N.inverse(d))

# Iteration controller
ic = ift.GradientNormController(iteration_limit=50000, tol_abs_gradnorm=0.1)
D_inv = S.inverse + R.adjoint @ N.inverse @ R
# Enable .inverse to invert D via Conjugate Gradient.
D_inv = ift.InversionEnabler(D_inv, ic)
D = D_inv.inverse

# + [markdown] slideshow={"slide_type": "subslide"}
# ### Apply Wiener Filter
#
# After defining the information source and propagator, we are able to apply the Wiener filter in order to get the posterior mean $m = \langle s \rangle_{\mathcal{P}(s|d)}$ that is our reconstruction of the signal:

# + slideshow={"slide_type": "-"}
m = D(j)

# + [markdown] slideshow={"slide_type": "subslide"}
# ### Results

# + slideshow={"slide_type": "-"}
# `.asnumpy()` retrieves the underlying numpy array from a NIFTy Field.
plt.plot(s.asnumpy(), 'r', label="signal ground truth", linewidth=2)
plt.plot(d.asnumpy(), 'k.', label="noisy data")
plt.plot(m.asnumpy(), 'k', label="posterior mean",linewidth=2)
plt.title("Reconstruction")
plt.legend()
plt.show()
# -

# To show the deviations with respect to the true signal (or ground truth), we  plot the residuals as follows:

# + slideshow={"slide_type": "subslide"}
plt.plot(s.asnumpy() - s.asnumpy(), 'r', label="ground truth ref [$s-s$]", linewidth=2)
plt.plot(d.asnumpy() - s.asnumpy(), 'k.', label="noise [$d-Rs$]")
plt.plot(m.asnumpy() - s.asnumpy(), 'k', label="posterior mean - ground truth",linewidth=2)
plt.axhspan(-noise_amplitude,noise_amplitude, facecolor='0.9', alpha=.5)
plt.title("Residuals")
plt.legend()
plt.show()

# + [markdown] slideshow={"slide_type": "slide"}
# ## Wiener Filter on Incomplete Data
#
# Now we consider a case that the data is not complete.
# This might be the case in real situations as the instrument might not be able to receive data.
# In order to apply the Wiener filter to this case, we first need to build the response corresponding to the incomplete measurement in NIFTy!
# -

# ### Incomplete Measuring / Masking
# We need to build a mask operator which cuts out all the unobserved parts of the signal.
# Lets assume that we first observe the signal for some time, but then something goes wrong with our instrument and we don't collect data for a while.
# After fixing the instrument we can collect data again.
# This means that data lives on an unstructured domain as there is data missing for the period of time $t_{\text{off}}$ when the instrument was offline.
# In order to implement this incomplete measurement we need to define a new response operator $R$ which masks the signal for the time $t_{\text{off}}$.
#

# + slideshow={"slide_type": "-"}
# Whole observation time
npix = s_space.size
# Time when the instrument is turned off
l = int(npix * 0.2)
# Time when the instrument is turned on again
h = int(npix * 0.4)
# Initialise a new array for the whole time frame
mask = np.zeros(s_space.shape, bool)
# Define the mask
mask[l:h] = 1
# Turn the numpy array into a nifty field
mask = ift.makeField(s_space, mask)
# Define the response operator which masks the places where mask == 1
R = ift.MaskOperator(mask)
# -

# ### Synthetic Data
# As in the Wiener filter example with complete data, we generate some synthetic data now by propagating the previously drawn prior sample through the incomplete measurement response and adding a noise sample.

# Define the noise covariance
N = ift.ScalingOperator(R.target, noise_amplitude**2, float)
# Draw a noise sample
n = N.draw_sample()
# Measure the signal sample with additional noise
d = R(s) + n

# ### Sampling from D
# Since we have an incomplete measurement we want to know how uncertain we are about our Wiener filter solution. We can easily obtain both, the mean and the standard deviation by sampling from $D$ and computing them directly from the drawn samples.
# In order to enable NIFTy to sample from $D$ we need to use some helper functions.

# + slideshow={"slide_type": "skip"}
# This implements the rule how to sample from a sum of covariances
D_inv = ift.SamplingEnabler(ift.SandwichOperator.make(cheese=N.inverse, bun=R), S.inverse, ic, S.inverse)
# Allow for numerical inversion
D_inv = ift.InversionEnabler(D_inv, ic)
D = D_inv.inverse
# Define the information source
j = R.adjoint(N.inverse(d))
# Posterior mean
m = D(j)

# Number of samples to calculate the posterior standard deviation
n_samples = 200

# Helper function that calculates the mean and the variance from a set of samples efficiently
sc = ift.StatCalculator()
for _ in range(n_samples):
    # Draw a sample of G(s,D) and shifting it by m -> G(s-m,D)
    sample = m + D.draw_sample()
    # Add it to the StatCalculator
    sc.add(sample)
# Standard deviation from samples
samples_std = sc.var.sqrt()
# Mean from samples that converges to m in the limit of infinitely many samples
samples_mean = sc.mean
# -

# ### Plots
# Let us visualize the results of the Wiener filter $m$, the sampled standard deviation and mean, as well as the true signal (ground truth) and the data.
# Since the data lives in data space, we first need to project it back into the signal space via $R^{\dagger}d$.

# + slideshow={"slide_type": "skip"}
plt.axvspan(l, h, facecolor='0.8',alpha=0.3, label="masked area") # Shading the masked interval

plt.plot(s.asnumpy(), '#f28109', label="Signal (ground truth)", alpha=1, linewidth=2) #
plt.plot(m.asnumpy(), 'k', label="Posterior mean (reconstruction)", linewidth=2)
plt.fill_between(range(m.size), (m - samples_std).asnumpy(), (m + samples_std).asnumpy(),
                 facecolor='#8592ff', alpha=0.8, label="Posterior std (samples)")
plt.plot(samples_mean.asnumpy(), 'k--', label="Posterior mean (samples)")

#.asnumpy() would return a read only-array. `.asnumpy_rw()` returns a writeable copy
data_s_space = R.adjoint(d).asnumpy_rw()
# Remove the "0" data points in the masked array
data_s_space[l:h] = np.nan
plt.plot(data_s_space, 'k.', label="Data")

plt.title("Reconstruction of incomplete data")
plt.legend()
plt.show()

# + [markdown] slideshow={"slide_type": "slide"}
# ## Wiener Filter standardized
# This problem can be solved in various coordinates and thus we can also choose the coordinate system, in which the prior $\mathcal{P}(s)=\mathcal{G}(0,\mathbb{1})$. And therefore the coordinate transformation from this to our original space is part of the prior description.
# +
# Coordinate transformation from Gaussian white noise to our prior from above (G(0,1)) to G(s,S)).
sqrt_pspec = S_k(ift.full(S_k.domain, 1.)).sqrt()
coord_trafo = HT.adjoint @ ift.makeOp(sqrt_pspec)

# The full Response from the standardized space to data space is a concatenation of R and the transformation
standardized_R = R @ coord_trafo
standardized_j = standardized_R.adjoint(N.inverse(d))

# standard_prior is now a indenty or unit matrix.
standardized_S = ift.Operator.identity_operator(standardized_R.domain)

# the new D
standardized_Dinv = ift.InversionEnabler(standardized_S + standardized_R.adjoint @ N.inverse @standardized_R, ic)
standardized_D = standardized_Dinv.inverse
standardized_m = standardized_D(standardized_j)
# -

# We can see that we get to the same result.

posterior_mean_s_space = coord_trafo(standardized_m)
plt.axvspan(l, h, facecolor='0.8',alpha=0.5)
plt.plot(s.asnumpy(), 'r', label="Signal", alpha=1, linewidth=2)
plt.plot(data_s_space, 'k.', label="Data")
plt.plot(posterior_mean_s_space.asnumpy(), 'k', label="Reconstruction", linewidth=2)
plt.title("Reconstruction of incomplete data in normalized coordinates")
plt.legend()
plt.show()

# Often it is easier to solve problems in this standardized parametere coordinates. For more information read:
#
# - Knollmüller, Jakob, and Torsten A. Enßlin. "Metric Gaussian variational inference." arXiv preprint arXiv:1901.11033 (2019).
# - Frank, Philipp, Reimar Leike, and Torsten A. Enßlin. "Geometric variational inference." Entropy 23.7 (2021): 853.



# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from typing import Callable, Optional, TypeVar, Union

import jax
from jax import numpy as jnp
from jax import random

from ..lax import fori_loop
from ..tree_math import ShapeWithDtype

V = TypeVar("V")


def lanczos_tridiag(mat: Callable[[V], V], v: V, order: int):
    """Compute the Lanczos decomposition into a tri-diagonal matrix and its
    corresponding orthonormal projection matrix.
    """
    swd = ShapeWithDtype.from_leave(v)
    tridiag = jnp.zeros((order, order), dtype=swd.dtype)
    vecs = jnp.zeros((order,) + swd.shape, dtype=swd.dtype)

    v = v / jnp.linalg.norm(v)
    vecs = vecs.at[0].set(v)

    # TODO
    # * use `tree_math.dot` and `tree_math.norm` in favor of plain `jnp.dot`
    # * remove all reshapes as they are unnecessary

    # Zeroth iteration
    w = mat(v)
    if w.shape != swd.shape:
        ve = f"shape of `mat(v)` {w.shape!r} incompatible with {swd}"
        raise ValueError(ve)
    alpha = jnp.dot(w, v)
    tridiag = tridiag.at[(0, 0)].set(alpha)
    w -= alpha * v
    beta = jnp.linalg.norm(w)

    tridiag = tridiag.at[(0, 1)].set(beta)
    tridiag = tridiag.at[(1, 0)].set(beta)
    vecs = vecs.at[1].set(w / beta)

    def reortho_step(j, state):
        vecs, w = state

        tau = vecs[j, :].reshape(swd.shape)
        coeff = jnp.dot(w, tau)
        w -= coeff * tau
        return vecs, w

    def lanczos_step(i, state):
        tridiag, vecs, beta = state

        # TODO: only save current and last vector and do not
        # reorthogonalize??????; check theory beforehand!!!
        v = vecs[i, :].reshape(swd.shape)
        v_old = vecs[i - 1, :].reshape(swd.shape)

        w = mat(v) - beta * v_old
        alpha = jnp.dot(w, v)
        tridiag = tridiag.at[(i, i)].set(alpha)
        w -= alpha * v

        # Full reorthogonalization
        # NOTE, in theory the loop could terminate at `i` but this would make
        # JAX's default backwards pass not work
        vecs, w = fori_loop(0, order, reortho_step, (vecs, w))

        # TODO: Raise if lanczos vectors are independent i.e. `beta` small?
        beta = jnp.linalg.norm(w)

        tridiag = tridiag.at[(i, i + 1)].set(beta)
        tridiag = tridiag.at[(i + 1, i)].set(beta)
        vecs = vecs.at[i + 1].set(w / beta)

        return tridiag, vecs, beta

    tridiag, vecs, beta = fori_loop(1, order - 1, lanczos_step, (tridiag, vecs, beta))

    # Final tridiag value and reorthogonalization
    v = vecs[order - 1, :].reshape(swd.shape)
    v_old = vecs[order - 2, :].reshape(swd.shape)
    w = mat(v) - beta * v_old
    alpha = jnp.dot(w, v)
    tridiag = tridiag.at[(order - 1, order - 1)].set(alpha)
    w -= alpha * v
    vecs, w = fori_loop(0, order - 1, reortho_step, (vecs, w))

    return (tridiag, vecs)


def stochastic_logdet_from_lanczos(
    tridiag_stack: jnp.ndarray, matrix_shape0: int, func: Callable = jnp.log
):
    """Computes a stochastic estimate of the log-determinate of a matrix using
    its Lanczos decomposition.

    Implemented via the stoachstic Lanczos quadrature.
    """
    eig_vals, eig_vecs = jnp.linalg.eigh(tridiag_stack)
    # TODO: Mask Eigenvalues <= 0?

    num_random_probes = tridiag_stack.shape[0]

    eig_ves_first_component = eig_vecs[..., 0, :]
    func_of_eig_vals = func(eig_vals)

    dot_products = jnp.sum(eig_ves_first_component**2 * func_of_eig_vals)
    return matrix_shape0 / float(num_random_probes) * dot_products


def stochastic_lq_logdet(
    mat: Union[jnp.ndarray, Callable],
    order: int,
    n_samples: int,
    key: Union[int, jnp.ndarray],
    *,
    shape0: Optional[int] = None,
    dtype=None,
    cmap=jax.vmap,
):
    """Computes a stochastic estimate of the log-determinate of a matrix using
    the stochastic Lanczos quadrature algorithm.
    """
    shape0 = shape0 if shape0 is not None else mat.shape[0]
    mat = mat.__matmul__ if not hasattr(mat, "__call__") else mat
    if not isinstance(key, jnp.ndarray):
        key = random.PRNGKey(key)
    key_smpls = random.split(key, n_samples)

    def random_lanczos(k):
        v = random.rademacher(k, (shape0,), dtype=dtype)
        tri, _ = lanczos_tridiag(mat, v, order=order)
        return tri

    tridiags = cmap(random_lanczos)(key_smpls)
    return stochastic_logdet_from_lanczos(tridiags, shape0)



# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from datetime import datetime
from functools import partial
from typing import Any, Callable, Dict, Mapping, NamedTuple, Optional, Tuple, Union

from jax import lax
from jax import numpy as jnp
from jax.tree_util import Partial

from . import conjugate_gradient
from .logger import logger
from .misc import (
    conditional_call,
    conditional_raise,
    doc_from,
    safeguard_arguments_against_accidental_calls_into_jax,
)
from .tree_math import (
    PyTreeString,
    assert_arithmetics,
    hide_strings,
    result_type,
    size,
    vdot,
    where,
)
from .tree_math import norm as jft_norm


class OptimizeResults(NamedTuple):
    """Object holding optimization results inspired by JAX and scipy.

    Attributes
    ----------
    x : ndarray
        The solution of the optimization.
    success : bool
        Whether or not the optimizer exited successfully.
    status : int
        Termination status of the optimizer. Its value depends on the
        underlying solver. NOTE, in contrast to scipy there is no `message` for
        details since strings are not statically memory bound.
    fun, jac, hess: ndarray
        Values of objective function, its Jacobian and its Hessian (if
        available). The Hessians may be approximations, see the documentation
        of the function in question.
    hess_inv : object
        Inverse of the objective function's Hessian; may be an approximation.
        Not available for all solvers.
    nfev, njev, nhev : int
        Number of evaluations of the objective functions and of its
        Jacobian and Hessian.
    nit : int
        Number of iterations performed by the optimizer.
    """

    x: Any
    success: Union[bool, jnp.ndarray]
    status: Union[int, jnp.ndarray]
    fun: Any
    jac: Any
    hess: Optional[jnp.ndarray] = None
    hess_inv: Optional[jnp.ndarray] = None
    nfev: Union[None, int, jnp.ndarray] = None
    njev: Union[None, int, jnp.ndarray] = None
    nhev: Union[None, int, jnp.ndarray] = None
    nit: Union[None, int, jnp.ndarray] = None
    # Trust-Region specific slots
    trust_radius: Union[None, float, jnp.ndarray] = None
    jac_magnitude: Union[None, float, jnp.ndarray] = None
    good_approximation: Union[None, bool, jnp.ndarray] = None


def _prepare_vag_hessp(fun, jac, hessp, fun_and_grad) -> Tuple[Callable, Callable]:
    """Returns a tuple of functions for computing the value-and-gradient and
    the Hessian-Vector-Product.
    """
    from warnings import warn

    if fun_and_grad is None:
        if fun is not None and jac is not None:
            uw = "computing the function together with its gradient would be faster"
            warn(uw, UserWarning)

            def fun_and_grad(x):
                return (fun(x), jac(x))

        elif fun is not None:
            from jax import value_and_grad

            fun_and_grad = value_and_grad(fun)
        else:
            ValueError("no function specified")

    if hessp is None:
        from jax import grad, jvp

        jac = grad(fun) if jac is None else jac

        def hessp(primals, tangents):
            return jvp(jac, (primals,), (tangents,))[1]

    return fun_and_grad, hessp


def newton_cg(fun=None, x0=None, *args, **kwargs):
    """Minimize a scalar-valued function using the Newton-CG algorithm."""
    if x0 is not None:
        assert_arithmetics(x0)
    return _newton_cg(fun, x0, *args, **kwargs).x


@doc_from(newton_cg)
def static_newton_cg(fun=None, x0=None, *args, **kwargs):
    if x0 is not None:
        assert_arithmetics(x0)
    return _static_newton_cg(fun, x0, *args, **kwargs).x


@safeguard_arguments_against_accidental_calls_into_jax
def _ncg_pretty_print_it(
    name,
    i,
    *,
    energy,
    energy_diff,
    grad_scaling,
    ls_reset,
    nhev,
    descent_norm,
    xtol,
    absdelta=None,
):
    msg = (
        f"{name}: →:{grad_scaling} ↺:{ls_reset} #∇²:{nhev:02d}"
        f" |↘|:{descent_norm:.6e} 🞋:{xtol:.6e}"
        f"\n{name}: Iteration {i} ⛰:{energy:+.6e} Δ⛰:{energy_diff:.6e}"
        + (f" 🞋:{absdelta:.6e}" if absdelta is not None else "")
    )
    logger.info(msg)


def _newton_cg(
    fun=None,
    x0=None,
    *,
    miniter=None,
    maxiter=None,
    energy_reduction_factor=0.1,
    old_fval=None,
    absdelta=None,
    norm_ord=None,
    xtol=1e-5,
    jac: Optional[Callable] = None,
    fun_and_grad=None,
    hessp=None,
    name=None,
    cg=conjugate_gradient._cg,
    cg_kwargs=None,
    time_threshold=None,
    custom_gradnorm=None,
):
    norm_ord = 1 if norm_ord is None else norm_ord
    miniter = 0 if miniter is None else miniter
    maxiter = 200 if maxiter is None else maxiter
    maxiter = jnp.iinfo(jnp.uint64).max if jnp.isinf(maxiter) else maxiter
    xtol = xtol * size(x0)

    pos = x0
    fun_and_grad, hessp = _prepare_vag_hessp(fun, jac, hessp, fun_and_grad=fun_and_grad)
    cg_kwargs = {} if cg_kwargs is None else cg_kwargs
    cg_name = name + "CG" if name is not None else None

    pp = partial(_ncg_pretty_print_it, name, xtol=xtol, absdelta=absdelta)

    gradnorm = (
        partial(jft_norm, ord=norm_ord) if custom_gradnorm is None else custom_gradnorm
    )
    energy, g = fun_and_grad(pos)
    nfev, njev, nhev = 1, 1, 0
    if jnp.isnan(energy):
        raise ValueError("energy is Nan")
    status = -1
    i = 0
    for i in range(1, maxiter + 1):
        # Newton approximates the potential up to second order. The CG energy
        # (`0.5 * x.T @ A @ x - x.T @ b`) and the approximation to the true
        # potential in Newton thus live on comparable energy scales. Hence, the
        # energy in a Newton minimization can be used to set the CG energy
        # convergence criterion.
        if old_fval and energy_reduction_factor:
            cg_absdelta = energy_reduction_factor * (old_fval - energy)
        else:
            cg_absdelta = None if absdelta is None else absdelta / 100.0
        mag_g = jft_norm(g, ord=cg_kwargs.get("norm_ord", 1))
        cg_resnorm = jnp.minimum(0.5, jnp.sqrt(mag_g)) * mag_g  # taken from SciPy
        default_kwargs = {
            "absdelta": cg_absdelta,
            "resnorm": cg_resnorm,
            "norm_ord": 1,
            "_raise_nonposdef": False,  # handle non-pos-def
            "name": cg_name,
            "time_threshold": time_threshold,
        }
        cg_res = cg(Partial(hessp, pos), g, **{**default_kwargs, **cg_kwargs})
        nat_g, info = cg_res.x, cg_res.info
        nhev += cg_res.nfev
        if info is not None and info < 0:
            raise ValueError("conjugate gradient failed")

        naive_ls_it = 0
        dd = nat_g  # negative descent direction
        grad_scaling = 1.0
        ls_reset = False
        for naive_ls_it in range(9):
            new_pos = pos - grad_scaling * dd
            new_energy, new_g = fun_and_grad(new_pos)
            nfev, njev = nfev + 1, njev + 1
            if new_energy <= energy:
                break

            grad_scaling /= 2
            if naive_ls_it == 5:
                ls_reset = True
                gam = float(vdot(g, g))
                curv = float(vdot(g, hessp(pos, g)))
                nhev += 1
                grad_scaling = 1.0
                dd = gam / curv * g
        else:
            grad_scaling = 0.0
            nm = "N" if name is None else name
            msg = f"{nm}: WARNING: Energy would increase; aborting"
            logger.warning(msg)
            status = -1
            break

        energy_diff = energy - new_energy
        old_fval = energy
        energy = new_energy
        pos = new_pos
        g = new_g

        descent_norm = grad_scaling * gradnorm(dd)
        if name is not None:
            pp(
                i,
                energy=energy,
                energy_diff=energy_diff,
                grad_scaling=grad_scaling,
                ls_reset=ls_reset,
                nhev=nhev,
                descent_norm=descent_norm,
            )
        if jnp.isnan(new_energy):
            raise ValueError("energy is NaN")
        min_cond = naive_ls_it < 2 and i > miniter
        if absdelta is not None and 0.0 <= energy_diff < absdelta and min_cond:
            status = 0
            break
        if descent_norm <= xtol and i > miniter:
            status = 0
            break
        if time_threshold is not None and datetime.now() > time_threshold:
            status = i
            break
    else:
        status = i
        nm = "N" if name is None else name
        logger.error(f"{nm}: Iteration Limit Reached")
    return OptimizeResults(
        x=pos,
        success=True,
        status=status,
        fun=energy,
        jac=g,
        nit=i,
        nfev=nfev,
        njev=njev,
        nhev=nhev,
    )


def _static_newton_cg(
    fun=None,
    x0=None,
    *,
    miniter=None,
    maxiter=None,
    energy_reduction_factor=0.1,
    old_fval=None,
    absdelta=None,
    norm_ord=None,
    xtol=1e-5,
    jac: Optional[Callable] = None,
    fun_and_grad=None,
    hessp=None,
    cg=conjugate_gradient._static_cg,
    name=None,
    cg_kwargs=None,
    custom_gradnorm=None,
):
    from jax.debug import callback

    from .lax import while_loop

    norm_ord = 1 if norm_ord is None else norm_ord
    miniter = 0 if miniter is None else miniter
    maxiter = 200 if maxiter is None else maxiter
    xtol = xtol * size(x0)

    pos = x0
    fun_and_grad, hessp = _prepare_vag_hessp(fun, jac, hessp, fun_and_grad=fun_and_grad)
    cg_kwargs = {} if cg_kwargs is None else cg_kwargs
    cg_name = name + "CG" if name is not None else None

    nm = "N" if name is None else name

    pp = partial(
        callback, _ncg_pretty_print_it, hide_strings(name), xtol=xtol, absdelta=absdelta
    )

    gradnorm = (
        partial(jft_norm, ord=norm_ord) if custom_gradnorm is None else custom_gradnorm
    )
    energy, g = fun_and_grad(pos)
    conditional_raise(jnp.isnan(energy), ValueError("energy is Nan"))
    val = {
        "status": jnp.where(maxiter == 0, 0, -2),
        "iteration": 0,
        "pos": pos,
        "energy": energy,
        "old_energy": old_fval if old_fval is not None else jnp.inf,
        "g": g,
        "nfev": 1,
        "njev": 1,
        "nhev": 0,
    }

    def continue_condition_newton_cg(v):
        return v["status"] < -1

    def single_newton_cg_step(v):
        status, i = v["status"], v["iteration"] + 1
        pos = v["pos"]
        energy, g = v["energy"], v["g"]
        old_energy = v["old_energy"]
        nfev, njev, nhev = v["nfev"], v["njev"], v["nhev"]

        cg_absdelta = 0.0 if absdelta is None else absdelta / 100.0
        if energy_reduction_factor is not None:
            cg_absdelta = jnp.where(
                ~jnp.isinf(old_energy),
                energy_reduction_factor * (old_energy - energy),
                cg_absdelta,
            )
        cg_absdelta = jnp.array(cg_absdelta, dtype=energy.dtype)
        mag_g = jft_norm(g, ord=cg_kwargs.get("norm_ord", 1))
        cg_resnorm = jnp.minimum(0.5, jnp.sqrt(mag_g)) * mag_g  # taken from SciPy
        default_kwargs = {
            "absdelta": cg_absdelta,
            "resnorm": cg_resnorm,
            "norm_ord": 1,
            "_raise_nonposdef": False,  # handle non-pos-def
            "name": cg_name,
        }
        cg_res = cg(Partial(hessp, pos), g, **{**default_kwargs, **cg_kwargs})
        nat_g, info = cg_res.x, cg_res.info
        nhev += cg_res.nfev
        conditional_raise(
            (info is not None) & (info < 0), ValueError("conjugate Gradient failed")
        )

        ret_ls = _line_search_successive_halving(
            pos, energy, g, nat_g, fun_and_grad, hessp, nm
        )
        status = jnp.where(ret_ls["status"] != 0, -1, status)

        # only update values if line search was successful
        old_energy = jnp.where(status < -1, energy, old_energy)
        energy = jnp.where(status < -1, ret_ls["new_energy"], energy)
        energy_diff = jnp.where(status < -1, old_energy - energy, 0.0)
        pos = where(status < -1, ret_ls["new_pos"], pos)
        g = where(status < -1, ret_ls["new_g"], g)

        grad_scaling = jnp.where(status < -1, ret_ls["grad_scaling"], 0.0)

        nfev += ret_ls["nfev"]
        njev += ret_ls["njev"]
        nhev += ret_ls["nhev"]

        descent_norm = grad_scaling * gradnorm(ret_ls["dd"])
        if name is not None:
            pp(
                i=i,
                grad_scaling=grad_scaling,
                ls_reset=ret_ls["reset"],
                nhev=nhev,
                descent_norm=descent_norm,
                energy=energy,
                energy_diff=energy_diff,
            )
        status = jnp.where(jnp.isnan(energy), -1, status)
        conditional_raise(jnp.isnan(energy), ValueError("energy is NaN"))
        min_cond = (ret_ls["iteration"] < 2) & (i > miniter)
        status = jnp.where(
            (0.0 <= energy_diff)
            & (False if absdelta is None else (energy_diff < absdelta))
            & min_cond
            & (status != -1),
            0,
            status,
        )
        status = jnp.where(
            (descent_norm <= xtol) & (i > miniter) & (status != -1), 0, status
        )
        status = jnp.where((i == maxiter) & (status < -1), i, status)

        ret = {
            "status": status,
            "iteration": i,
            "pos": pos,
            "energy": energy,
            "old_energy": old_energy,
            "g": g,
            "nfev": nfev,
            "njev": njev,
            "nhev": nhev,
        }
        return ret

    val = while_loop(continue_condition_newton_cg, single_newton_cg_step, val)
    conditional_call(
        (val["status"] > 0) | (maxiter == 0),
        logger.error,
        PyTreeString(f"{nm}: Iteration Limit Reached!"),
    )
    return OptimizeResults(
        x=val["pos"],
        success=True,
        status=val["status"],
        fun=val["energy"],
        jac=val["g"],
        nit=val["iteration"],
        nfev=val["nfev"],
        njev=val["njev"],
        nhev=val["nhev"],
    )


def _line_search_successive_halving(
    pos, start_energy, g, nat_g, fun_and_grad, hessp, name
):
    from .lax import cond, while_loop

    val = {
        "status": -2,
        "iteration": 0,
        "new_pos": pos,  # placeholder value
        "new_energy": jnp.inf,  # placeholder value
        "new_g": g,  # placeholder value
        "dd": nat_g,  # negative descent direction
        "grad_scaling": 1.0,
        "reset": False,
        "nfev": 0,
        "njev": 0,
        "nhev": 0,
    }

    def continue_condition_line_search(val):
        return val["status"] < -1

    def line_search_single_step(val):
        status = val["status"]
        i, reset = val["iteration"], val["reset"]
        grad_scaling, dd = val["grad_scaling"], val["dd"]
        nfev, njev, nhev = val["nfev"], val["njev"], val["nhev"]

        new_pos = pos - grad_scaling * dd
        new_energy, new_g = fun_and_grad(new_pos)
        nfev, njev = nfev + 1, njev + 1

        status = jnp.where(new_energy <= start_energy, 0, status)
        grad_scaling = jnp.where(status < -1, grad_scaling / 2, grad_scaling)

        do_reset = (i == 5) & (status < -1)
        reset = jnp.where(do_reset, True, reset)
        grad_scaling = jnp.where(do_reset, 1.0, grad_scaling)
        dd = cond(
            do_reset,
            lambda x: vdot(g, g) / vdot(g, hessp(pos, g)) * g,
            lambda x: x,
            dd,
        )
        nhev += do_reset

        do_abort = (i == 8) & (status < -1)
        status = jnp.where(do_abort, -1, status)
        conditional_call(
            do_abort,
            logger.error,
            PyTreeString(
                f"{name}: WARNING: Energy would increase; aborting line search."
            ),
        )

        ret = {
            "status": status,
            "iteration": i + 1,
            "new_pos": new_pos,
            "new_energy": new_energy,
            "new_g": new_g,
            "dd": dd,
            "grad_scaling": grad_scaling,
            "reset": reset,
            "nfev": nfev,
            "njev": njev,
            "nhev": nhev,
        }
        return ret

    return while_loop(continue_condition_line_search, line_search_single_step, val)


class _TrustRegionState(NamedTuple):
    x: Any
    converged: Union[bool, jnp.ndarray]
    status: Union[int, jnp.ndarray]
    fun: Any
    jac: Any
    nfev: Union[int, jnp.ndarray]
    njev: Union[int, jnp.ndarray]
    nhev: Union[int, jnp.ndarray]
    nit: Union[int, jnp.ndarray]
    trust_radius: Union[float, jnp.ndarray]
    jac_magnitude: Union[float, jnp.ndarray]
    old_fval: Union[float, jnp.ndarray]


def _trust_ncg(
    fun=None,
    x0=None,
    *,
    maxiter: Optional[int] = None,
    energy_reduction_factor=0.1,
    old_fval=jnp.nan,
    absdelta=None,
    gtol: float = 1e-4,
    max_trust_radius: Union[float, jnp.ndarray] = 1000.0,
    initial_trust_radius: Union[float, jnp.ndarray] = 1.0,
    eta: Union[float, jnp.ndarray] = 0.15,
    subproblem=conjugate_gradient._cg_steihaug_subproblem,
    jac: Optional[Callable] = None,
    hessp: Optional[Callable] = None,
    fun_and_grad: Optional[Callable] = None,
    subproblem_kwargs: Optional[Dict[str, Any]] = None,
    name: Optional[str] = None,
) -> OptimizeResults:
    from jax.debug import callback

    maxiter = 200 if maxiter is None else maxiter

    status = jnp.where(maxiter == 0, 1, 0)

    if not (0 <= eta < 0.25):
        raise Exception("invalid acceptance stringency")
    # Exception("gradient tolerance must be positive")
    status = jnp.where(gtol < 0.0, -1, status)
    # Exception("max trust radius must be positive")
    status = jnp.where(max_trust_radius <= 0, -1, status)
    # ValueError("initial trust radius must be positive")
    status = jnp.where(initial_trust_radius <= 0, -1, status)
    # ValueError("initial trust radius must be less than the max trust radius")
    status = jnp.where(initial_trust_radius >= max_trust_radius, -1, status)

    common_dtp = result_type(x0)
    eps = 6.0 * jnp.finfo(common_dtp).eps  # Inspired by SciPy's NewtonCG minimzer

    fun_and_grad, hessp = _prepare_vag_hessp(fun, jac, hessp, fun_and_grad=fun_and_grad)
    subproblem_kwargs = {} if subproblem_kwargs is None else subproblem_kwargs
    cg_name = name + "SP" if name is not None else None

    f_0, g_0 = fun_and_grad(x0)
    g_0_mag = jft_norm(g_0, ord=subproblem_kwargs.get("norm_ord", 1))
    status = jnp.where(jnp.isfinite(g_0_mag), status, 2)
    init_params = _TrustRegionState(
        converged=False,
        status=status,
        nit=0,
        x=x0,
        fun=f_0,
        jac=g_0,
        jac_magnitude=g_0_mag,
        nfev=1,
        njev=1,
        nhev=0,
        trust_radius=initial_trust_radius,
        old_fval=old_fval,
    )

    @safeguard_arguments_against_accidental_calls_into_jax
    def pp(arg):
        i = arg["i"]
        msg = (
            "{name}: ↗:{tr:.6e} ⬤:{hit} ∝:{rho:.2e} #∇²:{nhev:02d}"
            "\n{name}: Iteration {i} ⛰:{energy:+.6e} Δ⛰:{energy_diff:.6e}"
            + (" ➽:{absdelta:.6e}" if absdelta is not None else "")
            + ("\n{name}: Iteration Limit Reached" if i == maxiter else "")
        )
        logger.info(msg.format(name=name, **arg))

    def _trust_region_body_f(params: _TrustRegionState) -> _TrustRegionState:
        x_k, g_k, g_k_mag = params.x, params.jac, params.jac_magnitude
        i, f_k, old_fval = params.nit, params.fun, params.old_fval
        tr = params.trust_radius

        i += 1

        if energy_reduction_factor:
            cg_absdelta = energy_reduction_factor * (old_fval - f_k)
        else:
            cg_absdelta = None if absdelta is None else absdelta / 100.0
        cg_resnorm = jnp.minimum(0.5, jnp.sqrt(g_k_mag)) * g_k_mag
        # TODO: add an internal success check for future subproblem approaches
        # that might not be solvable
        default_kwargs = {
            "absdelta": cg_absdelta,
            "resnorm": cg_resnorm,
            "trust_radius": tr,
            "norm_ord": 1,
            "name": cg_name,
        }
        sub_result = subproblem(
            f_k, g_k, Partial(hessp, x_k), **{**default_kwargs, **subproblem_kwargs}
        )

        pred_f_kp1 = sub_result.pred_f
        x_kp1 = x_k + sub_result.step
        f_kp1, g_kp1 = fun_and_grad(x_kp1)

        actual_reduction = f_k - f_kp1
        pred_reduction = f_k - pred_f_kp1

        # update the trust radius according to the actual/predicted ratio
        rho = actual_reduction / pred_reduction
        tr_kp1 = jnp.where(rho < 0.25, tr * 0.25, tr)
        tr_kp1 = jnp.where(
            (rho > 0.75) & sub_result.hits_boundary,
            jnp.minimum(2.0 * tr, max_trust_radius),
            tr_kp1,
        )

        # compute norm to check for convergence
        g_kp1_mag = jft_norm(g_kp1, ord=subproblem_kwargs.get("norm_ord", 1))

        # if the ratio is high enough then accept the proposed step
        f_kp1, x_kp1, g_kp1, g_kp1_mag = where(
            rho > eta, (f_kp1, x_kp1, g_kp1, g_kp1_mag), (f_k, x_k, g_k, g_k_mag)
        )

        # Check whether we arrived at the float precision
        energy_eps = eps * jnp.abs(f_kp1)
        converged = (actual_reduction <= energy_eps) & (actual_reduction > -energy_eps)

        converged |= g_kp1_mag < gtol
        if absdelta:
            converged |= (
                (rho > eta) & (actual_reduction > 0.0) & (actual_reduction < absdelta)
            )

        status = jnp.where(converged, 0, params.status)
        status = jnp.where(i >= maxiter, 1, status)
        status = jnp.where(pred_reduction <= 0, 2, status)
        params = _TrustRegionState(
            converged=converged,
            nit=i,
            x=x_kp1,
            fun=f_kp1,
            jac=g_kp1,
            jac_magnitude=g_kp1_mag,
            nfev=params.nfev + sub_result.nfev + 1,
            njev=params.njev + sub_result.njev + 1,
            nhev=params.nhev + sub_result.nhev,
            trust_radius=tr_kp1,
            status=status,
            old_fval=f_k,
        )
        if name is not None:
            printable_state = {
                "i": i,
                "energy": params.fun,
                "energy_diff": actual_reduction,
                "maxiter": maxiter,
                "absdelta": absdelta,
                "tr": params.trust_radius,
                "rho": rho,
                "nhev": params.nhev,
                "hit": sub_result.hits_boundary,
            }
            callback(pp, printable_state)
        return params

    def _trust_region_cond_f(params: _TrustRegionState) -> bool:
        return jnp.logical_not(params.converged) & (params.status == 0)

    state = lax.while_loop(_trust_region_cond_f, _trust_region_body_f, init_params)

    return OptimizeResults(
        success=state.converged & (state.status == 0),
        nit=state.nit,
        x=state.x,
        fun=state.fun,
        jac=state.jac,
        nfev=state.nfev,
        njev=state.njev,
        nhev=state.nhev,
        jac_magnitude=state.jac_magnitude,
        trust_radius=state.trust_radius,
        status=state.status,
    )


def trust_ncg(fun=None, x0=None, *args, **kwargs):
    if x0 is not None:
        assert_arithmetics(x0)
    return _trust_ncg(fun, x0, *args, **kwargs).x


def minimize(
    fun: Optional[Callable[..., float]],
    x0,
    args: Tuple = (),
    *,
    method: str,
    tol: Optional[float] = None,
    options: Optional[Mapping[str, Any]] = None,
) -> OptimizeResults:
    """Minimize fun."""
    assert_arithmetics(x0)
    if options is None:
        options = {}
    if not isinstance(args, tuple):
        te = f"args argument must be a tuple, got {type(args)!r}"
        raise TypeError(te)

    fun_with_args = fun
    if args:
        fun_with_args = lambda x: fun(x, *args)

    if tol is not None:
        raise ValueError("use solver-specific options")

    if method.lower() in ("newton-cg", "newtoncg", "ncg"):
        return _newton_cg(fun_with_args, x0, **options)
    elif method.lower() in ("trust-ncg", "trustncg"):
        return _trust_ncg(fun_with_args, x0, **options)

    raise ValueError(f"method {method} not recognized")



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2022 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

############################################################
# Proof of concept of the evidence lower bound (ELBO)
# for model comparison.
#
# Conceptually the model is simply a 1D version of
# getting_started_3.py.
# The signal is a sigmoid-normal distributed field.
# The data is the field integrated along lines of sight that
# are randomly (set mode=0) or radially (mode=1) distributed,
# and it is drawn from a process with model 1 statistics.
#
# The data is finally fit with both model 1 and model 2.
# The respective evidences (ELBOs) are compared in order to
# establish which model is to be preferred.
#
# Demo takes a while to compute (especially the ELBO part).
#############################################################

import sys

import nifty.cl as ift
import numpy as np

ift.random.push_sseq_from_seed(28)


def random_los(n_los):
    starts = list(ift.random.current_rng().random((n_los, 1)).T)
    ends = list(ift.random.current_rng().random((n_los, 1)).T)
    return starts, ends


def radial_los(n_los):
    starts = list(ift.random.current_rng().random((n_los, 1)).T)
    ends = list(0.5 + 0 * ift.random.current_rng().random((n_los, 1)).T)
    return starts, ends


def main():
    # Choose between random line-of-sight response (mode=0) and radial lines
    # of sight (mode=1)
    if len(sys.argv) == 2:
        mode = int(sys.argv[1])
    else:
        mode = 0
    filename = "getting_started_model_comparison_mode_{}_".format(mode) + "{}.png"
    position_space = ift.RGSpace(128)

    #  For a detailed showcase of the effects the parameters
    #  of the CorrelatedField model have on the generated fields,
    #  see 'getting_started_4_CorrelatedFields.ipynb'.

    args_1 = {
        'offset_mean': 0,
        'offset_std': (1e-3, 1e-6),

        # Amplitude of field fluctuations
        'fluctuations': (1., 0.8),  # 1.0, 1e-2

        # Exponent of power law power spectrum component
        'loglogavgslope': (-3., 1),  # -6.0, 1

        # Amplitude of integrated Wiener process power spectrum component
        'flexibility': (2, 1.),  # 1.0, 0.5

        # How ragged the integrated Wiener process component is
        'asperity': (0.5, 0.4)  # 0.1, 0.5
    }

    args_2 = {
        'offset_mean': 0,
        'offset_std': (1e-3, 1e-6),

        # Amplitude of field fluctuations
        'fluctuations': (3., 0.8),  # 1.0, 1e-2

        # Exponent of power law power spectrum component
        'loglogavgslope': (-3., 1),  # -6.0, 1

        # Amplitude of integrated Wiener process power spectrum component
        'flexibility': (1.5, 1.),  # 1.0, 0.5

        # How ragged the integrated Wiener process component is
        'asperity': (2.5, 0.4)  # 0.1, 0.5
    }

    correlated_field_1 = ift.SimpleCorrelatedField(position_space, **args_1)
    pspec_1 = correlated_field_1.power_spectrum

    correlated_field_2 = ift.SimpleCorrelatedField(position_space, **args_2)
    pspec_2 = correlated_field_2.power_spectrum

    # Apply a nonlinearity
    signal_1 = ift.sigmoid(correlated_field_1)
    signal_2 = ift.sigmoid(correlated_field_2)

    # Build the line-of-sight response and define signal response
    LOS_starts, LOS_ends = random_los(100) if mode == 0 else radial_los(100)
    R = ift.LOSResponse(position_space, starts=LOS_starts, ends=LOS_ends)
    signal_response_1 = R(signal_1)
    signal_response_2 = R(signal_2)

    # Specify noise
    data_space = R.target
    noise = .001
    N = ift.ScalingOperator(data_space, noise, np.float64)

    # Generate mock signal and data
    mock_position = ift.from_random(signal_response_1.domain, 'normal')
    data = signal_response_1(mock_position) + N.draw_sample()

    # Minimization parameters
    ic_sampling = ift.AbsDeltaEnergyController(name="Sampling (linear)",
                                               deltaE=0.05, iteration_limit=100)
    ic_newton = ift.AbsDeltaEnergyController(name='Newton', deltaE=0.5,
                                             convergence_level=2, iteration_limit=35)
    ic_sampling_nl = ift.AbsDeltaEnergyController(name='Sampling (nonlin)',
                                                  deltaE=0.5, iteration_limit=15,
                                                  convergence_level=2)
    minimizer = ift.NewtonCG(ic_newton)
    minimizer_sampling = ift.NewtonCG(ic_sampling_nl)

    # Set up likelihood energy
    likelihood_energy_1 = (ift.GaussianEnergy(data=data, inverse_covariance=N.inverse) @
                           signal_response_1)

    likelihood_energy_2 = (ift.GaussianEnergy(data=data, inverse_covariance=N.inverse) @
                           signal_response_2)

    # Minimize KL
    n_iterations = 6
    n_samples = lambda iiter: 10 if iiter < 5 else 20
    samples_1 = ift.optimize_kl(likelihood_energy_1, n_iterations, n_samples, minimizer, ic_sampling,
                                nonlinear_sampling_minimizer=minimizer_sampling,
                                export_operator_outputs={"signal": signal_1},
                                output_directory="getting_started_model_comparison_results/model_1")
    samples_2 = ift.optimize_kl(likelihood_energy_2, n_iterations, n_samples, minimizer, ic_sampling,
                                nonlinear_sampling_minimizer=minimizer_sampling,
                                export_operator_outputs={"signal": signal_2},
                                output_directory="getting_started_model_comparison_results/model_2")

    # Plotting
    filename_res = filename.format("results")
    plot = ift.Plot()
    mean_1, var_1 = samples_1.sample_stat(signal_1)
    mean_2, var_2 = samples_2.sample_stat(signal_2)
    plot.add(data, title="Data")
    plot.add([signal_1(mock_position), mean_1, mean_2], title='Reconstruction',
             label=['Ground truth', 'Posterior Mean 1', 'Posterior Mean 2'])
    plot.add(var_1.sqrt(), title="Posterior Standard Deviation Model 1")
    plot.add(var_2.sqrt(), title="Posterior Standard Deviation Model 2")

    nsamples_1 = samples_1.n_samples
    logspec_1 = pspec_1.log()
    plot.add(list(samples_1.iterator(pspec_1)) +
             [pspec_1.force(mock_position), samples_1.average(logspec_1).exp()],
             title="Sampled Posterior Power Spectrum Model 1",
             linewidth=[1.] * nsamples_1 + [3., 3.],
             label=[None] * nsamples_1 + ['Ground truth', 'Posterior mean'])

    nsamples_2 = samples_2.n_samples
    logspec_2 = pspec_2.log()
    plot.add(list(samples_2.iterator(pspec_2)) +
             [pspec_2.force(mock_position), samples_2.average(logspec_2).exp()],
             title="Sampled Posterior Power Spectrum Model 2",
             linewidth=[1.] * nsamples_2 + [3., 3.],
             label=[None] * nsamples_2 + ['Ground truth (model 2)', 'Posterior mean'])

    plot.output(ny=2, nx=3, xsize=24, ysize=12, name=filename_res)
    ift.logger.info("Saved results as '{}'.".format(filename_res))

    # Compute evidence lower bound
    evidence_1, _ = ift.estimate_evidence_lower_bound(ift.StandardHamiltonian(likelihood_energy_1), samples_1, 100,
                                                      min_lh_eval=1e-3)
    evidence_2, _ = ift.estimate_evidence_lower_bound(ift.StandardHamiltonian(likelihood_energy_2), samples_2, 99,
                                                      min_lh_eval=1e-3)

    log_elbo_difference = evidence_1.average().asnumpy() - evidence_2.average().asnumpy()
    ift.logger.info("\n")
    elbo_ratio = np.exp(log_elbo_difference)
    if log_elbo_difference > 0:
        s = f"Model 1 is favored over model 2 by a factor of {elbo_ratio}."
    else:
        s = f"Model 2 is favored over model 1 by a factor of {1/elbo_ratio}."
    ift.logger.info(s)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# %% [markdown]
# # Nonlinear Models in NIFTy

# 1. Posterior: We would like to know $P(\theta|d)$ with $\theta$ the sky
#    brightness and $d$ measured count data of the sky brightness
# 1. Likelihood: We assume that $P(d|\theta)$ is a Poisson distribution
# 1. Prior: We assume that the sky brightness is a priori log-normal and
#    $\log \theta$ is spatially smooth
#
# To build a model in NIFTy, we go bottom and start from the prior, next we
# define the likelihood and finally retrieve (an approximation to) the posterior.

# %%
import nifty.cl as ift

# %% [markdown]
# ## Prior
#
# In NIFTy, we always start from a standard normal prior.
# Thus, instead of trying to directly create a smooth log-normal $\theta$, we
# instead ask ourselves (1) how we can make a standard normal smooth and
# afterwards (2) how we can make it log-normal.

# %%
position_space = ift.RGSpace([64, 64])  # domain on which our parameters live


# We need to apply the sqrt of the power spectrum to give a standard normal
# prior the desired power spectrum.
def power_spectrum_sqrt(k):
    return (1.0 / (20.0 + k**4))**0.5


p_space = ift.PowerSpace(position_space.get_default_codomain())
# Create an operator to distribute the power of the power-spectrum to indiviudal
# modes assuming the underlying field to be isotropic
pd = ift.PowerDistributor(position_space.get_default_codomain(), p_space)

a = ift.PS_field(p_space, power_spectrum_sqrt)
amplitude = pd(a)
amplitude = ift.makeOp(amplitude)
harmonic2pos = ift.HarmonicTransformOperator(amplitude.target, position_space)

# %%
r = ift.from_random(amplitude.domain)
ift.single_plot(harmonic2pos(amplitude(r)))

# %% [markdown]
# YAY, we achieved (1)

# %%
# Let's make it log-normal distributed. To do so we really only have to
# exponentiate it.
r = ift.from_random(amplitude.domain)
harmonic2pos = ift.HarmonicTransformOperator(amplitude.target, position_space)
ift.single_plot(ift.exp(harmonic2pos(amplitude(r))))

# %%
# We can also apply the operators to one another to retrieve a new operator that
# joins all of them. Here we create an operator to propagate our standard
# normally distributed prior parameters to smooth log-normal distributed
# parameter.
signal = ift.exp(harmonic2pos(amplitude))

# %% [markdown]
# YAY, we achieved (2)!

# %% [markdown]
# ## Likelihood
#
# We've done (1) and (2). Next, let us look at the likelihood $P(d|\theta)$.

# %%
# In any real life application, one would read in the actual data here. For
# simplicity, we synthetically create some data from our model.

# Create synthetic "true" latent parameters and propagate them through the model
r = ift.from_random(signal.domain)
synthetic_signal_realization = signal(r)
# Retrieve synthetic noisy data
rng = ift.random.current_rng()  # numpy random number generator
synthetic_data = rng.poisson(
    lam=synthetic_signal_realization.asnumpy(), size=position_space.shape
)
synthetic_data = ift.makeField(position_space, synthetic_data)

# %%
likelihood = ift.PoissonianEnergy(synthetic_data)

# %% [markdown]
# ## Posterior
#
# We now have our prior model and our likelihood model.
# Let's do some inference!

# %%
forward = likelihood @ signal
# NOTE, the optimization method only works with models that have named parameters.
# Thus, we need to give our parameters a name. This is usually not necessary
# for more complicated models (e.g. the correlated field model) as they are
# automatically assigned a name.
forward = forward.ducktape("domain")

ic_sampling = ift.DeltaEnergyController(
    name="Sampling", iteration_limit=200, tol_rel_deltaE=1e-8
)
ic_newton = ift.DeltaEnergyController(
    name="Newton", iteration_limit=35, tol_rel_deltaE=1e-8
)
minimizer = ift.NewtonCG(ic_newton)
# Increase this number (and/or the convergence criteria in `ic_*`) if you don't
# think your model converged yet
n_vi_iterations = 5
# Increase this number if you believe you got stuck in a weird local minimum
n_samples = 4

state = ift.optimize_kl(forward, n_vi_iterations, n_samples, minimizer, ic_sampling)

# %%
posterior_signal_samples = [
    signal.ducktape("domain")(sample) for sample in state.iterator()
]
p = ift.Plot()
p.add(synthetic_data, title="Synthetic Data")
for i in range(3):  # Show the first three samples
    p.add(posterior_signal_samples[i], title=f"Sample {i+1:02d}")
p.output()

# %%
p = ift.Plot()
m, v = state.sample_stat(signal.ducktape("domain"))
s = v**0.5
p.add(synthetic_signal_realization, title="Synthetic Signal Realization")
p.add(m, title="Posterior Mean")
p.add(s, title="Posterior Standard Deviation")
p.add(s / m, title="Posterior Relative Uncertainty")
p.output()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
# Authors: Gordian Edenhofer, Philipp Frank
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import string

import numpy as np

from ..domain_tuple import DomainTuple
from ..field import Field
from ..linearization import Linearization
from ..multi_domain import MultiDomain
from ..multi_field import MultiField
from ..utilities import myassert
from .linear_operator import LinearOperator
from .operator import Operator


class MultiLinearEinsum(Operator):
    """Multi-linear Einsum operator with corresponding derivates

    Parameters
    ----------
    domain : MultiDomain or dict{name: DomainTuple}
        The operator's input domain.
    subscripts : str
        The subscripts which is passed to einsum.
    key_order: tuple of str, optional
        The order of the keys in the multi-field. If not specified, defaults to
        the order of the keys in the multi-field.
    static_mf: MultiField or dict{name: Field}, optional
        A dictionary like type from which Fields are to be taken if the key from
        `key_order` is not part of the `domain`. Fields in this object are
        supposed to be static as they will not appear as FieldAdapter in the
        Linearization.
    optimize: bool, String or List, optional
        Parameter passed on to einsum_path.

    Notes
    -----
    The indices of the tensors run along spaces, not the axes of the spaces.
    For example, a matrix requires a product of two spaces, not one space with
    the combined shapes.

    By convention :class:`MultiLinearEinsum` only performs operations with
    lower indices. Therefore no complex conjugation is performed on complex
    inputs. To achieve operations with upper/lower indices use
    :class:`PartialConjugate` before applying this operator.
    """
    def __init__(self, domain, subscripts,
                 key_order=None, static_mf=None, optimize='optimal'):
        self._domain = MultiDomain.make(domain)
        if key_order is None:
            self._key_order = tuple(self._domain.keys())
        else:
            self._key_order = key_order
        if static_mf is not None and key_order is None:
            ve = "`key_order` mus be specified if additional fields are munged"
            raise ValueError(ve)
        self._stat_mf = static_mf
        iss, oss, *rest = subscripts.split("->")
        iss_spl = iss.split(",")
        len_consist = len(self._key_order) == len(iss_spl)
        sscr_consist = all(o in iss for o in oss)
        if rest or not sscr_consist or "," in oss or not len_consist:
            raise ValueError(f"invalid subscripts specified; got {subscripts}")
        ve = f"invalid order of keys {self._key_order} for subscripts {subscripts}"
        shapes, numpy_subscripts, subscriptmap = {}, '', {}
        alphabet = list(string.ascii_lowercase)[::-1]
        for k, ss in zip(self._key_order, iss_spl):
            dom = self._domain[k] if k in self._domain.keys(
            ) else self._stat_mf[k].domain
            if len(dom) != len(ss):
                raise ValueError(ve)
            for i, a in enumerate(list(ss)):
                if a not in subscriptmap.keys():
                    subscriptmap[a] = [alphabet.pop() for _ in
                                       range(len(dom[i].shape))]
                numpy_subscripts += ''.join(subscriptmap[a])
            numpy_subscripts += ','
            shapes[k] = dom.shape
        numpy_subscripts = numpy_subscripts[:-1] + '->'
        dom_sscr = dict(zip(self._key_order, iss_spl))
        tgt = []
        for o in oss:
            k_hit = tuple(k for k, sscr in dom_sscr.items() if o in sscr)[0]
            dom_k_idx = dom_sscr[k_hit].index(o)
            if k_hit in self._domain.keys():
                tgt += [self._domain[k_hit][dom_k_idx]]
            else:
                if k_hit not in self._stat_mf.keys():
                    ve = f"{k_hit} is not in domain nor in static_mf"
                    raise ValueError(ve)
                tgt += [self._stat_mf[k_hit].domain[dom_k_idx]]
            numpy_subscripts += ''.join(subscriptmap[o])
        self._target = DomainTuple.make(tgt)

        numpy_iss, numpy_oss, *_ = numpy_subscripts.split("->")
        numpy_iss_spl = numpy_iss.split(",")

        self._sscr_endswith = dict()
        self._linpaths = dict()
        for k, (i, ss) in zip(self._key_order, enumerate(numpy_iss_spl)):
            left_ss_spl = (*numpy_iss_spl[:i], *numpy_iss_spl[i + 1:], ss)
            linpath = '->'.join((','.join(left_ss_spl), numpy_oss))

            plc = tuple(np.broadcast_to(np.nan, shapes[q]) for q in shapes if q != k)
            plc += (np.broadcast_to(np.nan, shapes[k]),)
            self._sscr_endswith[k] = linpath
            self._linpaths[k] = np.einsum_path(linpath, *plc, optimize=optimize)[0]
        if isinstance(optimize, list):
            path = optimize
        else:
            plc = (np.broadcast_to(np.nan, shapes[k]) for k in shapes)
            path = np.einsum_path(numpy_subscripts, *plc, optimize=optimize)[0]
        self._sscr = numpy_subscripts
        self._ein_kw = {"optimize": path}

    def apply(self, x):
        self._check_input(x)
        if isinstance(x, Linearization):
            val = x.val.val
        else:
            val = x.val
        v = (
            val[k] if k in val else self._stat_mf[k].val
            for k in self._key_order
        )
        res = np.einsum(self._sscr, *v, **self._ein_kw)

        if isinstance(x, Linearization):
            jac = None
            for wrt in self.domain.keys():
                plc = {
                    k: x.val[k] if k in x.val else self._stat_mf[k]
                    for k in self._key_order if k != wrt
                }
                mf_wo_k = MultiField.from_dict(plc)
                ss = self._sscr_endswith[wrt]
                # Use the fact that the insertion order in a dictionary is the
                # ordering of keys as to pass on `key_order`
                jac_k = LinearEinsum(
                    self.domain[wrt],
                    mf_wo_k,
                    ss,
                    key_order=tuple(plc.keys()),
                    optimize=self._linpaths[wrt],
                    _target=self._target,
                    _calling_as_lin=True
                ).ducktape(wrt)
                jac = jac + jac_k if jac is not None else jac_k
            return x.new(Field.from_raw(self.target, res), jac)
        return Field.from_raw(self.target, res)


class LinearEinsum(LinearOperator):
    """Linear Einsum operator with exactly one freely varying field

    Parameters
    ----------
    domain : Domain, DomainTuple or tuple of Domain
        The operator's input domain.
    mf : :class:`nifty.cl.multi_field.MultiField`
        The first part of the left-hand side of the einsum.
    subscripts : str
        The subscripts which is passed to einsum. Everything before the very
        last scripts before the '->' is treated as part of the fixed mulfi-
        field while the last scripts are taken to correspond to the freely
        varying field.
    key_order: tuple of str, optional
        The order of the keys in the multi-field. If not specified, defaults to
        the order of the keys in the multi-field.
    optimize: bool, String or List, optional
        Parameter passed on to einsum_path.

    Notes
    -----
    The indices of the tensors run along spaces, not the axes of the spaces.
    For example, a matrix requires a product of two spaces, not one space with
    the combined shapes.

    By convention :class:`LinearEinsum` only performs operations with
    lower indices. Therefore no complex conjugation is performed on complex
    inputs or mf. To achieve operations with upper/lower indices use
    :class:`PartialConjugate` before applying this operator.
    """
    def __init__(self, domain, mf, subscripts, key_order=None, optimize='optimal',
                 _target=None, _calling_as_lin=False):
        self._domain = DomainTuple.make(domain)
        if _calling_as_lin:
            self._init_wo_preproc(mf, subscripts, key_order, optimize, _target)
        else:
            self._mf = mf
            if key_order is None:
                _key_order = tuple(self._mf.domain.keys())
            else:
                _key_order = key_order
            self._ein_kw = {"optimize": optimize}
            iss, oss, *rest = subscripts.split("->")
            iss_spl = iss.split(",")
            sscr_consist = all(o in iss for o in oss)
            len_consist = len(_key_order) == len(iss_spl[:-1])
            if rest or not sscr_consist or "," in oss or not len_consist:
                raise ValueError(f"invalid subscripts specified; got {subscripts}")
            ve = f"invalid order of keys {_key_order} for subscripts {subscripts}"
            shapes, numpy_subscripts, subscriptmap = (), '', {}
            alphabet = list(string.ascii_lowercase)
            for k, ss in zip(_key_order, iss_spl[:-1]):
                dom = self._mf[k].domain
                if len(dom) != len(ss):
                    raise ValueError(ve)
                for i, a in enumerate(list(ss)):
                    if a not in subscriptmap.keys():
                        subscriptmap[a] = [alphabet.pop() for _ in
                                           range(len(dom[i].shape))]
                    numpy_subscripts += ''.join(subscriptmap[a])
                numpy_subscripts += ','
                shapes += (dom.shape,)
            if len(self._domain) != len(iss_spl[-1]):
                raise ValueError(ve)
            for i, a in enumerate(list(iss_spl[-1])):
                if a not in subscriptmap.keys():
                    subscriptmap[a] = [alphabet.pop() for _ in
                                       range(len(self._domain[i].shape))]
                numpy_subscripts += ''.join(subscriptmap[a])
            shapes += (self._domain.shape,)
            numpy_subscripts += '->'

            dom_sscr = dict(zip(_key_order, iss_spl[:-1]))
            dom_sscr[id(self)] = iss_spl[-1]
            tgt = []
            for o in oss:
                k_hit = tuple(k for k, sscr in dom_sscr.items() if o in sscr)[0]
                dom_k_idx = dom_sscr[k_hit].index(o)
                if k_hit in _key_order:
                    tgt += [self._mf.domain[k_hit][dom_k_idx]]
                else:
                    myassert(k_hit == id(self))
                    tgt += [self._domain[dom_k_idx]]
                numpy_subscripts += "".join(subscriptmap[o])
            _target = DomainTuple.make(tgt)
            self._sscr = numpy_subscripts

            if isinstance(optimize, list):
                path = optimize
            else:
                plc = (np.broadcast_to(np.nan, shp) for shp in shapes)
                path = np.einsum_path(numpy_subscripts, *plc, optimize=optimize)[0]
            self._init_wo_preproc(mf, numpy_subscripts, _key_order, path, _target)

    def _init_wo_preproc(self, mf, subscripts, keyorder, optimize, target):
        self._ein_kw = {"optimize": optimize}
        self._mf = mf
        self._sscr = subscripts
        self._key_order = keyorder
        self._target = target
        iss, oss, *_ = subscripts.split("->")
        iss_spl = iss.split(",")

        if len(iss_spl) == 1:
            self._adj_sscr = "->".join((oss, iss))
        else:
            adj_iss = ",".join((",".join(iss_spl[:-1]), oss))
            self._adj_sscr = "->".join((adj_iss, iss_spl[-1]))
        self._capability = self.TIMES | self.ADJOINT_TIMES

    def _device_preparation(self, x, mode):
        self._mf = self._mf.at(x.device_id)

    def apply(self, x, mode):
        self._check_input(x, mode)
        self._device_preparation(x, mode)
        if mode == self.TIMES:
            dom, ss, mf = self.target, self._sscr, self._mf
        else:
            dom, ss, mf = self.domain, self._adj_sscr, self._mf.conjugate()
        res = np.einsum(
            ss, *(mf[k].val for k in self._key_order), x.val,
            **self._ein_kw
        )
        return Field.from_raw(dom, res)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# Authors: Gordian Edenhofer
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from ..domain_tuple import DomainTuple
from ..domains.rg_space import RGSpace
from ..domains.unstructured_domain import UnstructuredDomain
from ..field import Field
from ..multi_domain import MultiDomain
from ..multi_field import MultiField
from .linear_operator import LinearOperator


class SliceOperator(LinearOperator):
    """Geometry preserving mask operator

    Takes a field, slices it into the desired shape and returns the values of
    the field in the sliced domain all while preserving the original distances.

    Parameters
    ----------
    domain : Domain, DomainTuple or tuple of Domain
        The operator's input domain.
    new_shape : tuple of tuples or integers, or None
        The shape of the target domain with None indicating to copy the shape
        of the original domain for this axis. For example ((10, 5), 100) for a
        DomainTuple with two entires, the first having shape (10, 5) and the
        second having shape 100
    center : bool, optional
        Whether to center the slice that is selected in the input field.
    preserve_dist: bool, optional
        Whether to preserve the distance of the input field.
    """
    def __init__(self, domain, new_shape, center=False, preserve_dist=True):
        self._domain = DomainTuple.make(domain)
        if len(new_shape) != len(self._domain):
            ve = (
                f"shape ({new_shape}) is incompatible with the shape of the"
                f" domain ({self._domain.shape})"
            )
            raise ValueError(ve)
        for i, shape in enumerate(new_shape):
            if len(np.atleast_1d(shape)) != len(self._domain[i].shape):
                ve = (
                    f"shape of subspace ({i}) is incompatible with the domain"
                )
                raise ValueError(ve)

        tgt = []
        slc_by_ax = []
        for i, d in enumerate(self._domain):
            if new_shape[i] is None or np.all(
                np.array(self._domain.shape[i]) == np.array(new_shape[i])
            ):
                tgt += [d]
            elif np.all(np.array(new_shape[i]) <= np.array(d.shape)):
                dom_kw = dict()
                if isinstance(d, RGSpace):
                    if preserve_dist:
                        dom_kw["distances"] = d.distances
                    dom_kw["harmonic"] = d.harmonic
                elif not isinstance(d, UnstructuredDomain):
                    # Some domains like HPSpace or LMSPace can not be sliced
                    ve = f"{d.__class__.__name__} can not be sliced"
                    raise ValueError(ve)
                tgt += [d.__class__(new_shape[i], **dom_kw)]
            else:
                ve = (
                    f"domain axes ({d}) is smaller than the target shape"
                    f"{new_shape[i]}"
                )
                raise ValueError(ve)

            if center:
                for j, n_pix in enumerate(np.atleast_1d(new_shape[i])):
                    slc_start = np.floor((d.shape[j] - n_pix) / 2.).astype(int)
                    slc_end = slc_start + n_pix
                    slc_by_ax += [slice(slc_start, slc_end)]
            else:
                for n_pix in np.atleast_1d(new_shape[i]):
                    slc_start = 0
                    slc_end = n_pix
                    slc_by_ax += [slice(slc_start, slc_end)]

        self._slc_by_ax = tuple(slc_by_ax)
        self._target = DomainTuple.make(tgt)
        self._capability = self.TIMES | self.ADJOINT_TIMES

    def apply(self, x, mode):
        self._check_input(x, mode)
        x = x.asnumpy()
        if mode == self.TIMES:
            res = x[self._slc_by_ax]
            return Field.from_raw(self.target, res)
        res = np.zeros(self.domain.shape, x.dtype)
        res[self._slc_by_ax] = x
        return Field.from_raw(self.domain, res)

    def __str__(self):
        ss = (
            f"{self.__class__.__name__}"
            f"({self.domain.shape} -> {self.target.shape})"
        )
        return ss


class SplitOperator(LinearOperator):
    """Split a single field into a multi-field

    Takes a field, selects the desired entries for each multi-field key and
    puts the result into a multi-field. Along sliced axis, the domain will
    be replaced by an UnstructuredDomain as no distance measures are preserved.

    Note, slices may intersect, i.e. slices may reference the same input
    multiple times if the `intersecting_slices` option is set. However, a
    single field in the output may not contain the same part of the input more
    than once.

    Parameters
    ----------
    domain : Domain, DomainTuple or tuple of Domain
        The operator's input domain.
    slices_by_key : dict{key: tuple of integers or None}
        The key-value pairs of which the values indicate the parts to be
        selected. The result will be a multi-field with the given keys as
        entries and the selected slices of the domain as values. `None`
        indicates to select the whole input along this axis.
    intersecting_slices : bool, optional
        Tells the operator whether slices may contain intersections. If true,
        the adjoint is constructed a little less efficiently.  Set this
        parameter to `False` to gain a little more efficiency.
    """
    def __init__(self, domain, slices_by_key, intersecting_slices=True):
        self._domain = DomainTuple.make(domain)
        self._intersec_slc = intersecting_slices

        tgt = dict()
        self._k_slc = dict()
        for k, slc in slices_by_key.items():
            if len(slc) > len(self._domain):
                ve = f"slice at key {k!r} has more dimensions than the input"
                raise ValueError(ve)
            k_tgt = []
            k_slc_by_ax = []
            for i, d in enumerate(self._domain):
                if i >= len(slc) or slc[i] is None or (
                    isinstance(slc[i], slice) and slc[i] == slice(None)
                ):
                    k_tgt += [d]
                    k_slc_by_ax += [slice(None)]
                elif isinstance(slc[i], slice):
                    start = slc[i].start if slc[i].start is not None else 0
                    stop = slc[i].stop if slc[i].stop is not None else d.size
                    step = slc[i].step if slc[i].step is not None else 1
                    frac = np.floor((stop - start) / np.abs(step))
                    k_tgt += [UnstructuredDomain(frac.astype(int))]
                    k_slc_by_ax += [slc[i]]
                elif isinstance(slc[i],
                                np.ndarray) and slc[i].dtype is np.dtype(bool):
                    if slc[i].size != d.size:
                        raise ValueError(
                            "shape mismatch between desired slice {slc[i]}"
                            "and the shape of the domain {d.size}"
                        )
                    k_tgt += [UnstructuredDomain(slc[i].sum())]
                    k_slc_by_ax += [slc[i]]
                elif isinstance(slc[i], (tuple, list, np.ndarray)):
                    k_tgt += [UnstructuredDomain(len(slc[i]))]
                    k_slc_by_ax += [slc[i]]
                elif isinstance(slc[i], int):
                    k_slc_by_ax += [slc[i]]
                else:
                    ve = f"invalid type for specifying a slice; got {slc[i]}"
                    raise ValueError(ve)
            tgt[k] = DomainTuple.make(k_tgt)
            self._k_slc[k] = tuple(k_slc_by_ax)

        self._target = MultiDomain.make(tgt)
        self._capability = self.TIMES | self.ADJOINT_TIMES

    def apply(self, x, mode):
        self._check_input(x, mode)
        if mode == self.TIMES:
            res = dict()
            for k, slc in self._k_slc.items():
                res[k] = x.val[slc]
            return MultiField.from_raw(self.target, res)

        refarr = next(iter(x.val.values()))
        res = np.zeros_like(refarr, shape=self._tgt(mode).shape, dtype=refarr.dtype)
        if self._intersec_slc:
            for k, slc in self._k_slc.items():
                # Mind the `+` here for coping with intersections
                res[slc] += x.val[k]
            return Field.from_raw(self.domain, res)
        for k, slc in self._k_slc.items():
            res[slc] = x.val[k]
        return Field.from_raw(self.domain, res)

    def __str__(self):
        return f"{self.__class__.__name__} {self._target.keys()!r} <-"



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2022 Max-Planck-Society
# Copyright(C) 2022 Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

############################################################
# Similar to getting_started_3.py but with config files
#############################################################

import sys

import nifty.cl as ift


def sky(
    *,
    npix,
    vol,
    offset_mean,
    offset_std_mean,
    offset_std_std,
    fluctuations_mean,
    fluctuations_std,
    loglogavgslope_mean,
    loglogavgslope_std,
    flexibility_mean,
    flexibility_std,
    asperity_mean,
    asperity_std
):
    position_space = ift.RGSpace(npix, vol / npix)
    correlated_field = ift.SimpleCorrelatedField(
        position_space,
        offset_mean,
        offset_std=(offset_std_mean, offset_std_std),
        fluctuations=(fluctuations_mean, fluctuations_std),
        flexibility=(flexibility_mean, flexibility_std),
        asperity=(asperity_mean, asperity_std),
        loglogavgslope=(loglogavgslope_mean, loglogavgslope_std),
    )
    return correlated_field.sigmoid()


def lh(*, sky, noise_var):
    noise_var = float(noise_var)
    R, N, d = mock_RNd(
        sky, noise_var
    )  # This would be replaced by a data loading routine
    return ift.GaussianEnergy(d, inverse_covariance=N.inverse) @ R @ sky


def trans(*, sky_before, sky_after):
    # FIXME Write a transition that e.g. transitions from a lower to a higher resolved grid
    return lambda sample_list: ift.full(sky_after.domain, 0.)


builder_dct = {"lh0": lh, "lh1": lh, "sky0": sky, "sky1": sky, "trans01": trans}


def main():
    _, cfg_file = sys.argv
    cfg = ift.OptimizeKLConfig.from_file(cfg_file, builder_dct)
    cfg.optimize_kl(
        export_operator_outputs={"sky1": cfg.instantiate_section("sky1")},
        output_directory="output"
    )


def mock_RNd(signal_operator, noise_var):
    sdom = signal_operator.target
    p = 0.2
    with ift.random.Context(42):
        mask = ift.random.current_rng().choice(
            a=[False, True], size=sdom.shape, p=[p, 1 - p]
        )
        R = ift.MaskOperator(ift.makeField(sdom, mask))
        N = ift.ScalingOperator(R.target, noise_var, float)
        mock_position = ift.from_random(signal_operator.domain, "normal")
        d = (R @ signal_operator)(mock_position) + N.draw_sample()
    return R, N, d


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Copyright(C) 2013-2021 Max-Planck-Society
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

# %% [markdown]
# # Demonstration of a non-linear regression using NIFTy.re

# %%
import operator as op
from functools import partial

import jax
import matplotlib.pyplot as plt
import nifty.re as jft
import numpy as np
from jax import numpy as jnp
from jax import random

jax.config.update("jax_enable_x64", True)

# %%
seed = 42
key = random.PRNGKey(seed)


class NonLinearRegression(jft.Model):
    def __init__(self, slope_mean, slope_std, intercept_min, intercept_max, x):
        self.slope = jft.LogNormalPrior(slope_mean, slope_std, name="slope")
        self.intercept = jft.UniformPrior(
            intercept_min, intercept_max, name="intercept"
        )
        self.x = x
        super().__init__(init=self.slope.init | self.intercept.init)

    def __call__(self, xi, *, x=None):
        x = x if x is not None else self.x
        return x * self.slope(xi) + self.intercept(xi)


key, sk = random.split(key)
x = random.uniform(sk, (50,), float, -4.0, 4.0)
nlr = NonLinearRegression(3.0, 2.0, -5.0, 5.0, x)

# %%
noise_std = 5

key_tr, key_n = random.split(random.PRNGKey(31415))
y = nlr(nlr.init(key_tr)) + noise_std * random.normal(key_n, x.shape, x.dtype)

# %%
fig, ax = plt.subplots()
ax.plot(x, y, color="dodgerblue", linestyle="None", marker=".", markersize=8)
ax.set_xlabel("x")
ax.set_ylabel("y")
fig.tight_layout()
fig.savefig("nonlinear_regression_data.png")
plt.show()

# %%
lh = jft.Gaussian(y, noise_std_inv=partial(op.mul, 1.0 / noise_std)).amend(nlr)

# %%
key, ki, ko = random.split(key, 3)

delta = 1e-4
samples_opt, st = jft.optimize_kl(
    lh,
    jft.Vector(lh.init(ki)),
    key=ko,
    n_total_iterations=5,
    n_samples=12,
    draw_linear_kwargs=dict(
        cg_kwargs=dict(absdelta=delta * jft.size(lh.domain) / 10.0),
    ),
    nonlinearly_update_kwargs=dict(
        minimize_kwargs=dict(xtol=delta, cg_kwargs=dict(name=None))
    ),
    kl_kwargs=dict(
        minimize_kwargs=dict(name="M", xtol=delta, cg_kwargs=dict(name=None))
    ),
    sample_mode="nonlinear_resample",
)

# %%
x_p = np.linspace(x.min(), x.max(), 500)
y_unnoised_samples = jax.vmap(partial(nlr, x=x_p))(samples_opt.samples)

fig, ax = plt.subplots()
ax.plot(x, y, color="dodgerblue", linestyle="None", marker=".", markersize=8)
ax.plot(x_p, y_unnoised_samples.mean(axis=0), color="black")
qs = (0.16, 0.84)
ax.fill_between(
    x_p, *np.quantile(y_unnoised_samples, qs, axis=0), color="gray", alpha=0.3
)
ax.set_xlabel("x")
ax.set_ylabel("y")
fig.tight_layout()
fig.savefig("nonlinear_regression_posterior.png", dpi=400)
plt.show()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2022 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

############################################################
# Non-linear tomography
#
# The signal is a sigmoid-normal distributed field.
# The data is the field integrated along lines of sight that are
# randomly (set mode=0) or radially (mode=1) distributed
#
# Demo takes a while to compute
#############################################################

import sys

import numpy as np

import nifty.cl as ift

try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    master = comm.Get_rank() == 0
except ImportError:
    comm = None
    master = True


class SingleDomain(ift.LinearOperator):
    def __init__(self, domain, target):
        self._domain = ift.makeDomain(domain)
        self._target = ift.makeDomain(target)
        self._capability = self.TIMES | self.ADJOINT_TIMES

    def apply(self, x, mode):
        self._check_input(x, mode)
        return ift.makeField(self._tgt(mode), x.asnumpy())


def random_los(n_los):
    starts = list(ift.random.current_rng().random((n_los, 2)).T)
    ends = list(ift.random.current_rng().random((n_los, 2)).T)
    return starts, ends


def radial_los(n_los):
    starts = list(ift.random.current_rng().random((n_los, 2)).T)
    ends = list(0.5 + 0*ift.random.current_rng().random((n_los, 2)).T)
    return starts, ends


def main():
    # Choose between random line-of-sight response (mode=0) and radial lines
    # of sight (mode=1)
    if len(sys.argv) == 2:
        mode = int(sys.argv[1])
    else:
        mode = 0

    # Preparing the filename string for store results
    filename = "getting_started_mf_mode_{}_".format(mode) + "{}.png"

    # Set up signal domain
    npix1, npix2 = 128, 128
    position_space = ift.RGSpace([npix1, npix2])
    sp1 = ift.RGSpace(npix1)
    sp2 = ift.RGSpace(npix2)

    # Set up signal model
    cfmaker = ift.CorrelatedFieldMaker('')
    cfmaker.add_fluctuations(sp1, (0.1, 1e-2), (2, .2), (.01, .5), (-4, 2.),
                             'amp1')
    cfmaker.add_fluctuations(sp2, (0.1, 1e-2), (2, .2), (.01, .5), (-3, 1),
                             'amp2')
    cfmaker.set_amplitude_total_offset(0., (1e-2, 1e-6))
    correlated_field = cfmaker.finalize()

    normalized_amp = cfmaker.get_normalized_amplitudes()
    pspec1 = normalized_amp[0]**2
    pspec2 = normalized_amp[1]**2
    DC = SingleDomain(correlated_field.target, position_space)

    # Apply a nonlinearity
    signal = DC @ ift.sigmoid(correlated_field)

    # Build the line-of-sight response and define signal response
    LOS_starts, LOS_ends = random_los(100) if mode == 0 else radial_los(100)
    R = ift.LOSResponse(position_space, starts=LOS_starts, ends=LOS_ends)
    signal_response = R(signal)

    # Specify noise
    data_space = R.target
    noise = .001
    N = ift.ScalingOperator(data_space, noise, float)

    # Generate mock signal and data
    mock_position = ift.from_random(signal_response.domain, 'normal')
    data = signal_response(mock_position) + N.draw_sample()

    plot = ift.Plot()
    plot.add(signal(mock_position), title='Ground Truth')
    plot.add(R.adjoint_times(data), title='Data')
    plot.add([pspec1.force(mock_position)], title='Power Spectrum 1')
    plot.add([pspec2.force(mock_position)], title='Power Spectrum 2')
    plot.output(ny=2, nx=2, xsize=10, ysize=10, name=filename.format("setup"))

    # Minimization parameters
    ic_sampling = ift.AbsDeltaEnergyController(name='Sampling', deltaE=0.01, iteration_limit=100)
    ic_newton = ift.AbsDeltaEnergyController(name='Newton', deltaE=0.01, iteration_limit=35)
    minimizer = ift.NewtonCG(ic_newton, enable_logging=True)

    # Set up likelihood energy and information Hamiltonian
    likelihood_energy = ift.GaussianEnergy(data, inverse_covariance=N.inverse) @ signal_response

    n_samples = 20
    n_iterations = 5
    samples = ift.optimize_kl(likelihood_energy, n_iterations, n_samples,
                              minimizer, ic_sampling, comm=comm,
                              output_directory="getting_started_5_results",
                              export_operator_outputs={"signal": signal, "power spectrum 1": pspec1,
                                                       "power spectrum 2": pspec2})

    # Plotting
    filename_res = filename.format("results")
    plot = ift.Plot()
    mean, var = samples.sample_stat(signal)
    plot.add(mean, title="Posterior Mean")
    plot.add(ift.sqrt(var), title="Posterior Standard Deviation")

    n_samples = samples.n_samples
    plot.add(list(samples.iterator(pspec1)) + [samples.average(pspec1.log()).exp(),
              pspec1.force(mock_position)],
             title="Sampled Posterior Power Spectrum 1",
             linewidth=[1.]*n_samples + [3., 3.])
    plot.add(list(samples.iterator(pspec2)) + [samples.average(pspec2.log()).exp(),
              pspec2.force(mock_position)],
             title="Sampled Posterior Power Spectrum 2",
             linewidth=[1.]*n_samples + [3., 3.])
    if master:
        plot.output(ny=2, nx=2, xsize=15, ysize=15, name=filename_res)
    print("Saved results as '{}'.".format(filename_res))


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from functools import partial
from typing import Callable, Iterable, Optional, Tuple, Union

from jax import numpy as jnp
from jax import vmap

from ..misc import deprecated
from ..model import LazyModel
from ..multi_grid.kernel import refinement_matrices
from ..tree_math import Vector, sqrtm
from .chart import CoordinateChart
from .charted_refine import refine
from .util import RefinementMatrices, get_cov_from_loc, get_refinement_shapewithdtype


def _coordinate_pixel_refinement_matrices(
    chart: CoordinateChart,
    level: int,
    pixel_index: Optional[Iterable[int]] = None,
    kernel: Optional[Callable] = None,
    *,
    _cov_from_loc: Optional[Callable] = None,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    cov_from_loc = get_cov_from_loc(kernel, _cov_from_loc)
    csz = int(chart.coarse_size)  # coarse size
    if csz % 2 != 1:
        raise ValueError("only odd numbers allowed for `_coarse_size`")
    fsz = int(chart.fine_size)  # fine size
    if fsz % 2 != 0:
        raise ValueError("only even numbers allowed for `_fine_size`")
    ndim = chart.ndim
    if pixel_index is None:
        pixel_index = (0,) * ndim
    pixel_index = jnp.asarray(pixel_index)
    if pixel_index.size != ndim:
        ve = f"`pixel_index` has {pixel_index.size} dimensions but `chart` has {ndim}"
        raise ValueError(ve)

    csz_half = int((csz - 1) / 2)
    gc = jnp.arange(-csz_half, csz_half + 1, dtype=float)
    gc = jnp.ones((ndim, 1)) * gc
    gc = jnp.stack(jnp.meshgrid(*gc, indexing="ij"), axis=-1)
    if chart.fine_strategy == "jump":
        gf = jnp.arange(fsz, dtype=float) / fsz - 0.5 + 0.5 / fsz
    elif chart.fine_strategy == "extend":
        gf = jnp.arange(fsz, dtype=float) / 2 - 0.25 * (fsz - 1)
    else:
        raise ValueError(f"invalid `_fine_strategy`; got {chart.fine_strategy}")
    gf = jnp.ones((ndim, 1)) * gf
    gf = jnp.stack(jnp.meshgrid(*gf, indexing="ij"), axis=-1)
    # On the GPU a single `cov_from_loc` call is about twice as fast as three
    # separate calls for coarse-coarse, fine-fine and coarse-fine.
    coord = jnp.concatenate((gc.reshape(-1, ndim), gf.reshape(-1, ndim)), axis=0)
    del gc, gf
    coord = chart.ind2cart((coord + pixel_index.reshape((1, ndim))).T, level)
    coord = jnp.stack(coord, axis=-1)

    return refinement_matrices(cov_from_loc(coord, coord), fsz**ndim)


def _coordinate_refinement_matrices(
    chart: CoordinateChart,
    kernel: Callable,
    *,
    depth: Optional[int] = None,
    skip0=False,
    _cov_from_loc=None,
) -> RefinementMatrices:
    cov_from_loc = get_cov_from_loc(kernel, _cov_from_loc)
    depth = chart.depth if depth is None else depth

    if not skip0:
        rg0 = jnp.mgrid[tuple(slice(s) for s in chart.shape0)]
        c0 = jnp.stack(chart.ind2cart(rg0, 0), axis=-1).reshape(-1, chart.ndim)
        cov_sqrt0 = sqrtm(cov_from_loc(c0, c0))
    else:
        cov_sqrt0 = None

    opt_lin_filter, kernel_sqrt = [], []
    olf_at = vmap(
        partial(
            _coordinate_pixel_refinement_matrices, chart, _cov_from_loc=cov_from_loc
        ),
        in_axes=(None, 0),
        out_axes=(0, 0),
    )

    for lvl in range(depth):
        shape_lvl = chart.shape_at(lvl)
        pixel_indices = []
        for ax in range(chart.ndim):
            pad = (chart.coarse_size - 1) / 2
            if int(pad) != pad:
                raise ValueError("`coarse_size` must be odd")
            pad = int(pad)
            if chart.fine_strategy == "jump":
                stride = 1
            elif chart.fine_strategy == "extend":
                stride = chart.fine_size / 2
                if int(stride) != stride:
                    raise ValueError("`fine_size` must be even")
                stride = int(stride)
            else:
                raise AssertionError()
            if ax in chart.irregular_axes:
                pixel_indices.append(jnp.arange(pad, shape_lvl[ax] - pad, stride))
            else:
                pixel_indices.append(jnp.array([pad]))
        pixel_indices = jnp.stack(jnp.meshgrid(*pixel_indices, indexing="ij"), axis=-1)
        shape_filtered_lvl = pixel_indices.shape[:-1]
        pixel_indices = pixel_indices.reshape(-1, chart.ndim)

        olf, ks = olf_at(lvl, pixel_indices)
        shape_bc_lvl = tuple(
            shape_filtered_lvl[i] if i in chart.irregular_axes else 1
            for i in range(chart.ndim)
        )
        opt_lin_filter.append(olf.reshape(shape_bc_lvl + olf.shape[-2:]))
        kernel_sqrt.append(ks.reshape(shape_bc_lvl + ks.shape[-2:]))

    return RefinementMatrices(
        opt_lin_filter, kernel_sqrt, cov_sqrt0, (None,) * len(opt_lin_filter)
    )


class ChartedField(LazyModel):
    @deprecated("`ChartedField` is deprecated in favor of `ICRField`")
    def __init__(
        self,
        *args,
        kernel: Optional[Callable] = None,
        dtype=None,
        skip0: bool = False,
        **kwargs,
    ):
        """Initialize an Iterative Charted Refinement (ICR) field.

        There are multiple ways to initialize a charted refinement field. The
        recommended way is to first instantiate a `CoordinateChart` and pass it
        as first argument to this method. Alternatively, you may pass any and
        all arguments of `CoordinateChart` also to this method and it will
        instantiate the `CoordinateChart` for you and use it in the same way as
        if directly specified.

        Notes
        -----
        This method is deprecated, please use the new `ICRField` method.

        Parameters
        ----------
        chart : CoordinateChart
            The `CoordinateChart` with which to refine.
        kernel :
            Covariance kernel of the refinement field.
        dtype :
            Data-type of the excitations which to add during refining.
        skip0 :
            Whether to skip the first refinement level. This is useful to e.g.
            stack multiple refinement fields on top of each other.
        **kwargs :
            Alternatively to `chart` any parameters accepted by
            `CoordinateChart`.
        """
        self._kernel = kernel
        self._dtype = dtype
        self._skip0 = skip0

        if len(args) > 0 and isinstance(args[0], CoordinateChart):
            if kwargs:
                raise TypeError(f"expected no keyword arguments, got {kwargs}")

            if len(args) == 1:
                (self._chart,) = args
            elif len(args) == 2 and callable(args[1]) and kernel is None:
                self._chart, self._kernel = args
            elif (
                len(args) == 3
                and callable(args[1])
                and kernel is None
                and dtype is None
            ):
                self._chart, self._kernel, self._dtype = args
            elif (
                len(args) == 4
                and callable(args[1])
                and kernel is None
                and dtype is None
                and skip0 == False
            ):
                self._chart, self._kernel, self._dtype, self._skip0 = args
            else:
                te = "got unexpected arguments in addition to CoordinateChart"
                raise TypeError(te)
        else:
            self._chart = CoordinateChart(*args, **kwargs)

    @property
    def kernel(self):
        """Yields the kernel specified during initialization or throw a
        `TypeError`.
        """
        if self._kernel is None:
            te = (
                "either specify a fixed kernel during initialization of the"
                f" {self.__class__.__name__} class or provide one here"
            )
            raise TypeError(te)
        return self._kernel

    @property
    def dtype(self):
        """Yields the data-type of the excitations."""
        return jnp.float64 if self._dtype is None else self._dtype

    @property
    def skip0(self):
        """Whether to skip the zeroth refinement"""
        return self._skip0

    @property
    def chart(self):
        """Associated `CoordinateChart` with which to iterative refine."""
        return self._chart

    def matrices(
        self,
        kernel: Optional[Callable] = None,
        depth: Optional[int] = None,
        skip0: Optional[bool] = None,
        **kwargs,
    ) -> RefinementMatrices:
        """Computes the refinement matrices namely the optimal linear filter
        and the square root of the information propagator (a.k.a. the square
        root of the fine covariance matrix for the excitations) for all
        refinement levels and all pixel indices in the coordinate chart.

        Parameters
        ----------
        kernel :
            Covariance kernel of the refinement field if not specified during
            initialization.
        depth :
            Maximum refinement depth if different to the one of the `CoordinateChart`.
        skip0 :
            Whether to skip the first refinement level.
        """
        if kernel is None and "_cov_from_loc" not in kwargs:
            kernel = self.kernel
        depth = self.chart.depth if depth is None else depth
        skip0 = self.skip0 if skip0 is None else skip0

        return _coordinate_refinement_matrices(
            self.chart, kernel=kernel, depth=depth, skip0=skip0, **kwargs
        )

    def matrices_at(
        self,
        level: int,
        pixel_index: Optional[Iterable[int]] = None,
        kernel: Optional[Callable] = None,
        **kwargs,
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """Computes the refinement matrices namely the optimal linear filter
        and the square root of the information propagator (a.k.a. the square
        root of the fine covariance matrix for the excitations) at the
        specified level and pixel index.

        Parameters
        ----------
        level :
            Refinement level.
        pixel_index :
            Index of the NDArray at the refinement level `level` which to
            refine, i.e. use as center coarse pixel.
        kernel :
            Covariance kernel of the refinement field if not specified during
            initialization.
        """
        if kernel is None and "_cov_from_loc" not in kwargs:
            kernel = self.kernel

        return _coordinate_pixel_refinement_matrices(
            self.chart, level=level, pixel_index=pixel_index, kernel=kernel, **kwargs
        )

    @property
    def domain(self):
        """Yields the `ShapeWithDtype` of the primals."""
        return get_refinement_shapewithdtype(
            shape0=self.chart.shape0,
            depth=self.chart.depth,
            dtype=self.dtype,
            skip0=self.skip0,
            _coarse_size=self.chart.coarse_size,
            _fine_size=self.chart.fine_size,
            _fine_strategy=self.chart.fine_strategy,
        )

    @staticmethod
    def apply(
        xi,
        chart,
        kernel: Union[Callable, RefinementMatrices],
        *,
        skip0: bool = False,
        depth: Optional[int] = None,
        _refine: Optional[Callable] = None,
        _cov_from_loc: Optional[Callable] = None,
        precision=None,
    ):
        """Static method to apply a refinement field given some excitations, a
        chart and a kernel.

        Parameters
        ----------
        xi :
            Latent parameters which to use for refining.
        chart :
            Chart with which to refine.
        kernel :
            Covariance kernel with which to build the refinement matrices.
        skip0 :
            Whether to skip the first refinement level.
        depth :
            Refinement depth if different to the depth of the coordinate chart.
        precision :
            See JAX's precision.
        """
        depth = chart.depth if depth is None else depth
        if depth != len(xi.tree if isinstance(xi, Vector) else xi) - 1:
            ve = (
                f"incompatible refinement depths of `xi` ({len(xi) - 1})"
                f" and `depth` (of chart) {depth}"
            )
            raise ValueError(ve)

        if isinstance(kernel, RefinementMatrices):
            refinement = kernel
        else:
            refinement = _coordinate_refinement_matrices(
                chart,
                kernel=kernel,
                depth=depth,
                skip0=skip0,
                _cov_from_loc=_cov_from_loc,
            )
        refine_w_chart = partial(
            refine if _refine is None else _refine,
            _coarse_size=chart.coarse_size,
            _fine_size=chart.fine_size,
            _fine_strategy=chart.fine_strategy,
            precision=precision,
        )

        if not skip0:
            fine = (refinement.cov_sqrt0 @ xi[0].ravel()).reshape(xi[0].shape)
        else:
            if refinement.cov_sqrt0 is not None:
                raise AssertionError()
            fine = xi[0]
        for x, olf, k in zip(xi[1:], refinement.filter, refinement.propagator_sqrt):
            fine = refine_w_chart(fine, x, olf, k)
        return fine

    def __call__(self, xi, kernel=None, *, skip0=None, **kwargs):
        """See `ChartedField.apply`."""
        if kernel is None and "_cov_from_loc" not in kwargs:
            kernel = self.kernel
        skip0 = self.skip0 if skip0 is None else skip0
        return self.apply(xi, self.chart, kernel=kernel, skip0=skip0, **kwargs)

    def __repr__(self):
        descr = f"{self.__class__.__name__}({self.chart!r}"
        descr += f", kernel={self._kernel!r}" if self._kernel is not None else ""
        descr += f", dtype={self._dtype!r}" if self._dtype is not None else ""
        descr += f", skip0={self.skip0!r}" if self.skip0 is not False else ""
        descr += ")"
        return descr

    def __eq__(self, other):
        return repr(self) == repr(other)

    def __hash__(self):
        return hash(repr(self))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
# Copyright(C) 2025 Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from functools import reduce
from operator import add

import numpy as np
from scipy.sparse import coo_matrix

from ..domains.rg_space import RGSpace
from ..domains.unstructured_domain import UnstructuredDomain
from ..field import Field
from ..sugar import makeDomain
from .linear_operator import LinearOperator


class LinearInterpolator(LinearOperator):
    """Multilinear interpolation for points in an RGSpace

    Parameters
    ----------
    domain : RGSpace
    sampling_points : numpy.ndarray
        Positions at which to interpolate, shape (dim, ndata),

    Notes
    -----
    Positions that are not within the RGSpace are wrapped according to
    periodic boundary conditions. This reflects the general property of
    RGSpaces to be tori topologically.
    """
    def __init__(self, domain, sampling_points):
        self._domain = makeDomain(domain)
        for dom in self.domain:
            if not isinstance(dom, RGSpace):
                raise TypeError
        dims = [len(dom.shape) for dom in self.domain]

        # FIXME This needs to be removed as soon as the bug below is fixed.
        if dims.count(dims[0]) != len(dims):
            raise TypeError("This is a bug. Please extend"
                            "LinearInterpolator's functionality!")

        shp = sampling_points.shape
        if not (isinstance(sampling_points, np.ndarray) and len(shp) == 2):
            raise TypeError
        n_dim, n_points = shp
        if n_dim != reduce(add, dims):
            raise TypeError
        self._target = makeDomain(UnstructuredDomain(n_points))
        self._capability = self.TIMES | self.ADJOINT_TIMES
        self._build_mat(sampling_points, n_points)

    def _build_mat(self, sampling_points, N_points):
        ndim = sampling_points.shape[0]
        mg = np.mgrid[(slice(0, 2),)*ndim]
        mg = np.array(list(map(np.ravel, mg)))
        dist = [list(dom.distances) for dom in self.domain]
        # FIXME This breaks as soon as not all domains have the same number of
        # dimensions.
        dist = np.array(dist).reshape(-1, 1)
        pos = sampling_points/dist
        excess = pos - np.floor(pos)
        pos = np.floor(pos).astype(np.int64)
        max_index = np.array(self.domain.shape).reshape(-1, 1)
        data = np.zeros((len(mg[0]), N_points))
        ii = np.zeros((len(mg[0]), N_points), dtype=np.int64)
        jj = np.zeros((len(mg[0]), N_points), dtype=np.int64)
        for i in range(len(mg[0])):
            factor = np.prod(
                np.abs(1 - mg[:, i].reshape(-1, 1) - excess), axis=0)
            data[i, :] = factor
            fromi = (pos + mg[:, i].reshape(-1, 1)) % max_index
            ii[i, :] = np.arange(N_points)
            jj[i, :] = np.ravel_multi_index(fromi, self.domain.shape)
        self._mat = coo_matrix((data.reshape(-1),
                                (ii.reshape(-1), jj.reshape(-1))),
                               (N_points, np.prod(self.domain.shape)))
        self._finalize_init()

    def _finalize_init(self):
        if isinstance(self._mat, coo_matrix):
            from scipy.sparse.linalg import aslinearoperator
            self._device_id = -1
        else:
            from cupyx.scipy.sparse.linalg import aslinearoperator
            self._device_id = self._mat.data.device.id
        self._sop = aslinearoperator(self._mat)

    def _device_preparation(self, x, mode):
        if x.device_id == self._device_id:
            return
        if self._device_id > -1:  # need gpu->cpu copy
            self._mat = self._mat.get()
        elif self._device_id == -1:  # need cpu->gpu copy
            from cupyx.scipy.sparse import coo_matrix
            self._mat = coo_matrix(self._mat)
        else:
            raise RuntimeError()
        self._finalize_init()

    def apply(self, x, mode):
        self._check_input(x, mode)
        self._device_preparation(x, mode)
        x = x.raw
        if mode == self.TIMES:
            res = self._sop.matvec(x.reshape(-1))
        else:
            res = self._sop.rmatvec(x).reshape(self.domain.shape)
        return Field(self._tgt(mode), res)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from ..any_array import AnyArray
from ..domain_tuple import DomainTuple
from ..domains.rg_space import RGSpace
from ..field import Field
from ..utilities import infer_space, special_add_at
from .linear_operator import LinearOperator


class RegriddingOperator(LinearOperator):
    """Linearly interpolates an RGSpace to an RGSpace with coarser resolution.

    Parameters
    ----------
    domain : Domain, DomainTuple or tuple of Domain
        domain[space] needs to be an :class:`RGSpace`.
    new_shape : tuple of int
        Shape of the space which domain[space] is replaced by. Each entry must
        be smaller or equal to the respective entry in `domain[space].shape`.
    space : int
        Index of space in `domain` on which the operator shall act.
        Default is 0.
    """
    def __init__(self, domain, new_shape, space=0):
        self._domain = DomainTuple.make(domain)
        self._space = infer_space(self._domain, space)
        dom = self._domain[self._space]

        if not isinstance(dom, RGSpace):
            raise TypeError("RGSpace required")
        if len(new_shape) != len(dom.shape):
            raise ValueError("Shape mismatch")
        if any([a > b for a, b in zip(new_shape, dom.shape)]):
            raise ValueError("New shape must not be larger than old shape")
        if any([ii <= 0 for ii in new_shape]):
            raise ValueError('New shape must not be zero or negative.')

        newdist = tuple(dom.distances[i]*dom.shape[i]/new_shape[i]
                        for i in range(len(dom.shape)))

        tgt = RGSpace(new_shape, newdist)
        self._target = list(self._domain)
        self._target[self._space] = tgt
        self._target = DomainTuple.make(self._target)
        self._capability = self.TIMES | self.ADJOINT_TIMES

        ndim = len(new_shape)
        self._bindex = [None] * ndim
        self._frac = [None] * ndim
        for d in range(ndim):
            tmp = np.arange(new_shape[d])*(newdist[d]/dom.distances[d])
            self._bindex[d] = np.minimum(dom.shape[d]-2, tmp.astype(np.int64))
            self._frac[d] = tmp-self._bindex[d]
        self._bindex = [AnyArray(x) for x in self._bindex]
        self._frac = [AnyArray(x) for x in self._frac]

    def _device_preparation(self, x, mode):
        self._bindex = [elem.at(x.device_id) for elem in self._bindex]
        self._frac = [elem.at(x.device_id) for elem in self._frac]

    def apply(self, x, mode):
        self._check_input(x, mode)
        self._device_preparation(x, mode)
        v = x.val
        ndim = len(self.target.shape)
        curshp = list(self._dom(mode).shape)
        tgtshp = self._tgt(mode).shape
        d0 = self._target.axes[self._space][0]
        for d in self._target.axes[self._space]:
            idx = (slice(None),) * d
            wgt = self._frac[d-d0].reshape((1,)*d + (-1,) + (1,)*(ndim-d-1))

            if mode == self.ADJOINT_TIMES:
                shp = list(v.shape)
                shp[d] = tgtshp[d]
                xnew = np.zeros_like(v, shape=shp, dtype=v.dtype)
                xnew = special_add_at(xnew, d, self._bindex[d-d0], v*(1.-wgt))
                xnew = special_add_at(xnew, d, self._bindex[d-d0]+1, v*wgt)
            else:  # TIMES
                xnew = v[idx + (self._bindex[d-d0],)] * (1.-wgt)
                xnew += v[idx + (self._bindex[d-d0]+1,)] * wgt

            curshp[d] = xnew.shape[d]
            v = xnew
        return Field(self._tgt(mode), xnew)



# Copyright(C) 2013-2023 Gordian Edenhofer
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from .model import WrappedCall
from .num import (
    invgamma_prior,
    laplace_prior,
    lognormal_prior,
    normal_prior,
    uniform_prior,
)

_doc_shared = """name : hashable, optional
    Name within the new `input` on which `call` acts.
shape : tuple or tree-like structure of ShapeWithDtype
    Shape of the latent parameter(s) that are transformed to the desired
    distribution. This can also be an arbitrary shape-dtype structure in
    which case `dtype` is ignored. Defaults to a scalar.
dtype : dtype
    Data type of the latent parameter(s) that are transformed to the
    desired distribution."""


def _format_doc(func):
    func.__doc__ = func.__doc__.format(_doc_shared=_doc_shared)
    return func


class LaplacePrior(WrappedCall):
    @_format_doc
    def __init__(self, alpha, **kwargs):
        """Transforms standard normally distributed random variables to a
        Laplace distribution.

        Parameters
        ----------
        alpha : tree-like structure with arithmetics
            Scale parameter.
        {_doc_shared}
        """
        self.alpha = alpha
        call = laplace_prior(self.alpha)
        super().__init__(call, white_init=True, **kwargs)


class NormalPrior(WrappedCall):
    @_format_doc
    def __init__(self, mean, std, **kwargs):
        """Transforms standard normally distributed random variables to a
        normal distribution.

        Parameters
        ----------
        mean : tree-like structure with arithmetics
            Mean of the normal distribution.
        std : tree-like structure with arithmetics
            Standard deviation of the normal distribution.
        {_doc_shared}
        """
        self.mean = mean
        self.std = std
        call = normal_prior(self.mean, self.std)
        super().__init__(call, white_init=True, **kwargs)


class LogNormalPrior(WrappedCall):
    @_format_doc
    def __init__(self, mean, std, **kwargs):
        """Transforms standard normally distributed random variables to a
        log-normal distribution.

        Parameters
        ----------
        mean : tree-like structure with arithmetics
            Mean of the log-normal distribution.
        std : tree-like structure with arithmetics
            Standard deviation of the log-normal distribution.
        {_doc_shared}
        """
        self.mean = mean
        self.std = std
        call = lognormal_prior(self.mean, self.std)
        super().__init__(call, white_init=True, **kwargs)


class UniformPrior(WrappedCall):
    @_format_doc
    def __init__(self, a_min, a_max, **kwargs):
        """Transforms standard normally distributed random variables to a
        uniform distribution.

        Parameters
        ----------
        a_min : tree-like structure with arithmetics
            Minimum value.
        a_max : tree-like structure with arithmetics
            Maximum value.
        {_doc_shared}
        """
        self.low = self.a_min = a_min
        self.high = self.a_max = a_max
        call = uniform_prior(self.a_min, self.a_max)
        super().__init__(call, white_init=True, **kwargs)


class InvGammaPrior(WrappedCall):
    @_format_doc
    def __init__(self, a, scale, loc=0.0, step=1e-2, **kwargs):
        """Transforms standard normally distributed random variables to an
        inverse gamma distribution.

        Parameters
        ----------
        a : float
            Shape parameter.
        scale : float
            Scale parameter.
        loc : float
            Location parameter.
        step : float
            Step size for numerical integration.
        {_doc_shared}

        Notes
        -----
        Broadcasting over tree-like structure is not yet implemented. Please
        file an issue if you need this feature.
        """
        self.a = a
        self.scale = scale
        self.loc = loc
        self.step = step
        call = invgamma_prior(self.a, self.scale, self.loc, self.step)
        super().__init__(call, white_init=True, **kwargs)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from .. import utilities
from ..domain_tuple import DomainTuple
from ..domains.gl_space import GLSpace
from ..domains.lm_space import LMSpace
from ..domains.rg_space import RGSpace
from ..ducc_dispatch import fftn, hartley, ifftn
from ..field import Field
from .diagonal_operator import DiagonalOperator
from .linear_operator import LinearOperator
from .scaling_operator import ScalingOperator


class FFTOperator(LinearOperator):
    """Transforms between a pair of position and harmonic RGSpaces.

    Parameters
    ----------
    domain: Domain, tuple of Domain or DomainTuple
        The domain of the data that is input by "times" and output by
        "adjoint_times".
    target: Domain, optional
        The target (sub-)domain of the transform operation.
        If omitted, a domain will be chosen automatically.
    space: int, optional
        The index of the subdomain on which the operator should act
        If None, it is set to 0 if `domain` contains exactly one space.
        `domain[space]` must be an RGSpace.

    Notes
    -----
    This operator performs full FFTs, which implies that its output field will
    always have complex type, regardless of the type of the input field.
    If a real field is desired after a forward/backward transform couple, it
    must be manually cast to real.
    """

    def __init__(self, domain, target=None, space=None):
        # Initialize domain and target
        self._domain = DomainTuple.make(domain)
        self._capability = self._all_ops
        self._space = utilities.infer_space(self._domain, space)

        adom = self._domain[self._space]
        if not isinstance(adom, RGSpace):
            raise TypeError("FFTOperator only works on RGSpaces")
        if target is None:
            target = adom.get_default_codomain()

        self._target = [dom for dom in self._domain]
        self._target[self._space] = target
        self._target = DomainTuple.make(self._target)
        adom.check_codomain(target)
        target.check_codomain(adom)

    def apply(self, x, mode):
        self._check_input(x, mode)
        ncells = x.domain[self._space].size
        if x.domain[self._space].harmonic:  # harmonic -> position
            func = ifftn
            fct = ncells
        else:
            func = fftn
            fct = 1.
        axes = x.domain.axes[self._space]
        tdom = self._tgt(mode)
        tmp = func(x.val, axes=axes)
        Tval = Field(tdom, tmp)
        if mode & (LinearOperator.TIMES | LinearOperator.ADJOINT_TIMES):
            fct *= self._domain[self._space].scalar_dvol
        else:
            fct *= self._target[self._space].scalar_dvol
        return Tval if fct == 1 else Tval*fct


class HartleyOperator(LinearOperator):
    """Transforms between a pair of position and harmonic RGSpaces.

    Parameters
    ----------
    domain: Domain, tuple of Domain or DomainTuple
        The domain of the data that is input by "times" and output by
        "adjoint_times".
    target: Domain, optional
        The target (sub-)domain of the transform operation.
        If omitted, a domain will be chosen automatically.
    space: int, optional
        The index of the subdomain on which the operator should act
        If None, it is set to 0 if `domain` contains exactly one space.
        `domain[space]` must be an RGSpace.

    Notes
    -----
    This operator always produces output fields with the same data type as
    its input. This is achieved by performing so-called Hartley transforms
    (https://en.wikipedia.org/wiki/Discrete_Hartley_transform).
    For complex input fields, the operator will transform the real and
    imaginary parts separately and use the results as real and imaginary parts
    of the result field, respectively.
    In many contexts the Hartley transform is a perfect substitute for the
    Fourier transform, but in some situations (e.g. convolution with a general,
    non-symmetric kernel, the full FFT must be used instead.
    """

    def __init__(self, domain, target=None, space=None):
        # Initialize domain and target
        self._domain = DomainTuple.make(domain)
        self._capability = self._all_ops
        self._space = utilities.infer_space(self._domain, space)

        adom = self._domain[self._space]
        if not isinstance(adom, RGSpace):
            raise TypeError("HartleyOperator only works on RGSpaces")
        if target is None:
            target = adom.get_default_codomain()

        self._target = [dom for dom in self._domain]
        self._target[self._space] = target
        self._target = DomainTuple.make(self._target)
        adom.check_codomain(target)
        target.check_codomain(adom)

    def apply(self, x, mode):
        self._check_input(x, mode)
        if utilities.iscomplextype(x.dtype):
            return (self._apply_cartesian(x.real, mode) +
                    1j*self._apply_cartesian(x.imag, mode))
        else:
            return self._apply_cartesian(x, mode)

    def _apply_cartesian(self, x, mode):
        axes = x.domain.axes[self._space]
        tdom = self._tgt(mode)
        tmp = hartley(x.val, axes=axes)
        Tval = Field(tdom, tmp)
        if mode & (LinearOperator.TIMES | LinearOperator.ADJOINT_TIMES):
            fct = self._domain[self._space].scalar_dvol
        else:
            fct = self._target[self._space].scalar_dvol
        return Tval if fct == 1 else Tval*fct


class SHTOperator(LinearOperator):
    """Transforms between a harmonic domain on the sphere and a position
    domain counterpart.

    Built-in domain pairs are
      - an LMSpace and a HPSpace
      - an LMSpace and a GLSpace

    The supported operations are times() and adjoint_times().

    Parameters
    ----------
    domain : Domain, tuple of Domain or DomainTuple
        The domain of the data that is input by "times" and output by
        "adjoint_times".
    target : Domain, optional
        The target domain of the transform operation.
        If omitted, a domain will be chosen automatically.
        Whenever the input domain of the transform is an RGSpace, the codomain
        (and its parameters) are uniquely determined.
        For LMSpace, a GLSpace of sufficient resolution is chosen.
    space : int, optional
        The index of the domain on which the operator should act
        If None, it is set to 0 if domain contains exactly one subdomain.
        domain[space] must be a LMSpace.

    Note
    ----
    If this operator is called on inputs stored on device, it will copy them to
    host, perform the operations with the CPU and copy back.
    """

    def __init__(self, domain, target=None, space=None):
        # Initialize domain and target
        self._domain = DomainTuple.make(domain)
        self._capability = self.TIMES | self.ADJOINT_TIMES
        self._space = utilities.infer_space(self._domain, space)

        hspc = self._domain[self._space]
        if not isinstance(hspc, LMSpace):
            raise TypeError("SHTOperator only works on a LMSpace domain")
        if target is None:
            target = hspc.get_default_codomain()

        self._target = [dom for dom in self._domain]
        self._target[self._space] = target
        self._target = DomainTuple.make(self._target)
        hspc.check_codomain(target)
        target.check_codomain(hspc)

        from ducc0.sht import sharpjob_d
        self.lmax = hspc.lmax
        self.mmax = hspc.mmax
        self.sjob = sharpjob_d()
        self.sjob.set_triangular_alm_info(self.lmax, self.mmax)
        if isinstance(target, GLSpace):
            self.sjob.set_gauss_geometry(target.nlat, target.nlon)
        else:
            self.sjob.set_healpix_geometry(target.nside)

    def __reduce__(self):
        return (_unpickleSHTOperator,
                (self._domain, self._target[self._space], self._space))

    def apply(self, x, mode):
        self._check_input(x, mode)
        if utilities.iscomplextype(x.dtype):
            res = (self._apply_spherical(x.real, mode) +
                   1j*self._apply_spherical(x.imag, mode))
        else:
            res = self._apply_spherical(x, mode)
        return res.at(x.device_id)

    def _slice_p2h(self, inp):
        rr = self.sjob.alm2map_adjoint(inp)
        if len(rr) != ((self.mmax+1)*(self.mmax+2))//2 + \
                      (self.mmax+1)*(self.lmax-self.mmax):
            raise ValueError("array length mismatch")
        res = np.empty(2*len(rr)-self.lmax-1, dtype=rr[0].real.dtype)
        res[0:self.lmax+1] = rr[0:self.lmax+1].real
        res[self.lmax+1::2] = np.sqrt(2)*rr[self.lmax+1:].real
        res[self.lmax+2::2] = np.sqrt(2)*rr[self.lmax+1:].imag
        return res/np.sqrt(np.pi*4)

    def _slice_h2p(self, inp):
        res = np.empty((len(inp)+self.lmax+1)//2, dtype=(inp[0]*1j).dtype)
        if len(res) != ((self.mmax+1)*(self.mmax+2))//2 + \
                       (self.mmax+1)*(self.lmax-self.mmax):
            raise ValueError("array length mismatch")
        res[0:self.lmax+1] = inp[0:self.lmax+1]
        res[self.lmax+1:] = np.sqrt(0.5)*(inp[self.lmax+1::2] +
                                          1j*inp[self.lmax+2::2])
        res = self.sjob.alm2map(res)
        return res/np.sqrt(np.pi*4)

    def _apply_spherical(self, x, mode):
        axes = x.domain.axes[self._space]
        v = x.asnumpy()

        p2h = not x.domain[self._space].harmonic
        tdom = self._tgt(mode)
        func = self._slice_p2h if p2h else self._slice_h2p
        odat = np.empty(tdom.shape, dtype=x.dtype)
        for slice in utilities.get_slice_list(v.shape, axes):
            odat[slice] = func(v[slice])
        return Field(tdom, odat)


def _unpickleSHTOperator(*args):
    return SHTOperator(*args)


class HarmonicTransformOperator(LinearOperator):
    """Transforms between a harmonic domain and a position domain counterpart.

    Built-in domain pairs are
      - a harmonic and a non-harmonic RGSpace (with matching distances)
      - an LMSpace and a HPSpace
      - an LMSpace and a GLSpace

    The supported operations are times() and adjoint_times().
    If inverse_times() on RGSpaces is needed the HartleyOperator should be used instead.

    The harmonic transform on spheres (via `LMSpace`) is always computed on the
    cpu with an implicit copy to the host if the input resides on device.

    Parameters
    ----------
    domain : Domain, tuple of Domain or DomainTuple
        The domain of the data that is input by "times" and output by
        "adjoint_times".
    target : Domain, optional
        The target domain of the transform operation.
        If omitted, a domain will be chosen automatically.
        Whenever the input domain of the transform is an RGSpace, the codomain
        (and its parameters) are uniquely determined.
        For LMSpace, a GLSpace of sufficient resolution is chosen.
    space : int, optional
        The index of the domain on which the operator should act
        If None, it is set to 0 if domain contains exactly one subdomain.
        domain[space] must be a harmonic domain.

    Notes
    -----
    HarmonicTransformOperator uses a Hartley transformation to transform
    between harmonic and non-harmonic RGSpaces. This has the advantage that all
    field values are real in either space. If you require a true Fourier
    transform you should use FFTOperator instead.
    """

    def __init__(self, domain, target=None, space=None):
        domain = DomainTuple.make(domain)
        space = utilities.infer_space(domain, space)

        hspc = domain[space]
        if not hspc.harmonic:
            raise TypeError(
                "HarmonicTransformOperator only works on a harmonic space")
        if isinstance(hspc, RGSpace):
            self._op = HartleyOperator(domain, target, space)
        else:
            self._op = SHTOperator(domain, target, space)
        self._domain = self._op.domain
        self._target = self._op.target
        self._capability = self.TIMES | self.ADJOINT_TIMES

    def apply(self, x, mode):
        self._check_input(x, mode)
        return self._op.apply(x, mode)


def HarmonicSmoothingOperator(domain, sigma, space=None):
    """Returns an operator that carries out a smoothing with a Gaussian kernel
    of width `sigma` on the part of `domain` given by `space`.

    Parameters
    ----------
    domain : Domain, tuple of Domain, or DomainTuple
       The total domain of the operator's input and output fields
    sigma : float>=0
       The sigma of the Gaussian used for smoothing. It has the same units as
       the RGSpace the operator is working on.
       If `sigma==0`, an identity operator will be returned.
    space : int, optional
       The index of the sub-domain on which the smoothing is performed.
       Can be omitted if `domain` only has one sub-domain.

    Notes
    -----
    The sub-domain on which the smoothing is carried out *must* be a
    non-harmonic `RGSpace`.
    """

    sigma = float(sigma)
    if sigma < 0.:
        raise ValueError("sigma must be non-negative")
    if sigma == 0.:
        return ScalingOperator(domain, 1.)

    domain = DomainTuple.make(domain)
    space = utilities.infer_space(domain, space)
    if domain[space].harmonic:
        raise TypeError("domain must not be harmonic")
    Hartley = HartleyOperator(domain, space=space)
    codomain = Hartley.domain[space].get_default_codomain()
    kernel = codomain.get_k_length_array()
    smoother = codomain.get_fft_smoothing_kernel_function(sigma)
    kernel = smoother(kernel)
    ddom = list(domain)
    ddom[space] = codomain
    diag = DiagonalOperator(kernel, ddom, space)
    return Hartley.inverse(diag(Hartley))



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause
# Authors: Philipp Frank, Gordian Edenhofer

import operator
from collections import namedtuple
from functools import partial, reduce
from typing import Optional, Tuple

import jax.numpy as jnp
import numpy as np
from jax import eval_shape, jit, vmap
from jax.lax import scan
from numpy import typing as npt

from ..num import amend_unique_
from .grid import FlatGrid, Grid, OpenGridAtLevel
from .grid_impl import HEALPixGridAtLevel
from ..tree_math import solve, sqrtm


def apply_kernel(x, *, kernel):
    """Applies the kernel to values on an entire multigrid"""
    if len(x) != (kernel.grid.depth + 1):
        msg = f"input depth {len(x)} does not match grid depth {kernel.grid.depth}"
        raise ValueError(msg)
    for lvl, xx in enumerate(x):
        g = kernel.grid.at(lvl)
        if xx.size != g.size:
            msg = f"input at level {lvl} of size {xx.size} does not match grid of size {g.size}"
            raise ValueError(msg)

    def apply_at(index, level, x):
        assert index.ndim == 1
        iout, iin = kernel.get_output_input_indices(index, level)
        kernels = kernel.get_matrices(index, level)
        assert len(iin) == len(kernels)
        res = reduce(
            operator.add,
            (kk @ x[x_lvl][tuple(idx)] for kk, (idx, x_lvl) in zip(kernels, iin)),
        )
        return iout, res.reshape(iout[0].shape[1:])

    x = list(x)
    _, x[0] = apply_at(jnp.array([-1]), None, x)  # Use dummy index for base
    for lvl in range(kernel.grid.depth):
        g = kernel.grid.at(lvl)
        index = g.refined_indices()
        # TODO this selects window for each index individually
        f = apply_at
        for i in range(g.ndim):
            f = vmap(f, (1, None, None), ((g.ndim - i, None), g.ndim - i - 1))
        (_, lvl_nxt), res = f(index, lvl, x)
        x[lvl_nxt] = kernel.grid.at(lvl_nxt).resort(res)
    return x


_IdxMap = namedtuple("_IdxMap", ("shift", "index2flatindex"))
_CompressedIndexMap = namedtuple(
    "_CompressedIndexMap",
    ("base_kernel", "kernels", "uindices", "indexmaps", "invindices"),
)


class Kernel:
    """
    - Apply linear operators with inputs accross an arbitrary grid
    - A Kernel could be as simple as coarse graining children to a parent; or a full
      ICR or MSC implementation
    - Fully jax transformable to allow seamlessly being used in larger models
    """

    def __init__(self, grid, *, _cim: Optional[_CompressedIndexMap] = None):
        self._grid = grid
        self._cim = _cim

    def replace(self, *, _cim=None, **kwargs):
        _cim = self._cim if _cim is None else _cim
        return self.__class__(self.grid, **kwargs, _cim=_cim)

    @property
    def grid(self) -> Grid:
        return self._grid

    @property
    def compressed(self) -> bool:
        return self._cim is not None

    def get_output_input_indices(
        self, index, level: int
    ) -> Tuple[
        Tuple[npt.NDArray[np.int_], int], Tuple[Tuple[npt.NDArray[np.int_], int], ...]
    ]:
        raise NotImplementedError()

    def get_matrices(self, index, level):
        """Compute or lookup kernel matrices."""
        if self.compressed:
            return self.lookup_matrices(index, level)
        return self.compute_matrices(index, level)

    def compute_matrices(self, index, level):
        """Compute kernel matrices from scratch."""
        raise NotImplementedError()

    def lookup_matrices(self, index, level):
        """Efficient retrieval of kernel matrices for a compressed kernel."""
        if self._cim is None:
            msg = "kernel needs to be compressed first for fast lookups"
            raise NotImplementedError(msg)

        if level is None:
            return self._cim.base_kernel
        index = self._cim.indexmaps[level].index2flatindex(index)[0]
        index = self._cim.invindices[level][index - self._cim.indexmaps[level].shift]
        return tuple(kk[index] for kk in self._cim.kernels[level])

    def compress_indices(
        self,
        *,
        rtol=1e-5,
        atol=1e-10,
        buffer_size=10000,
        use_distances=True,
        distance_norm=partial(jnp.linalg.norm, axis=0),
    ):
        """Link kernel matrices to preserve memory and increase speed.

        You probably don't want to access this function directly. Instead use
        `compress` or `replace` and `compress_matrices`.
        """

        def get_distance_matrices(index, level):
            if level is None:
                raise NotImplementedError
            (out, olvl), ids = self.get_output_input_indices(index, level)
            out = out.reshape(index.shape + (-1,))
            out = self.grid.at(olvl).index2coord(out)
            assert index.ndim == out.ndim - 1
            ids = tuple(self.grid.at(ii[1]).index2coord(ii[0]) for ii in ids)
            ids = jnp.concatenate(ids, axis=-1)
            assert index.ndim == ids.ndim - 1
            return (distance_norm(out[..., jnp.newaxis] - ids[..., jnp.newaxis, :]),)

        gridf = self.grid if isinstance(self.grid, FlatGrid) else FlatGrid(self.grid)
        uindices = []
        invindices = []
        indexmaps = []
        for lvl in range(self.grid.depth):
            grid_at = self.grid.at(lvl)
            gridf_at = gridf.at(lvl)

            def get_matrices(idx):
                f = get_distance_matrices if use_distances else self.compute_matrices
                ker = f(gridf_at.flatindex2index(jnp.atleast_1d(idx)), lvl)
                ker = jnp.concatenate(tuple(kk.ravel() for kk in ker))
                return ker

            @jit
            def scanned_amend_unique(carry, idx, shift):
                (u, inv) = carry
                k = get_matrices(idx)
                u, invid = amend_unique_(u, k, axis=0, atol=atol, rtol=rtol)
                inv = inv.at[idx - shift].set(invid)
                return (u, inv), invid

            indices = grid_at.refined_indices()
            indices = gridf_at.index2flatindex(indices)[0].ravel()
            shift = np.min(indices)
            size = np.max(indices) - shift
            inv = np.full(size, buffer_size + 1)

            shp = eval_shape(get_matrices, indices[0]).shape
            unique = jnp.full((buffer_size,) + shp, jnp.nan)

            (unique, inv), invid = scan(
                partial(scanned_amend_unique, shift=shift), (unique, inv), indices
            )
            _, idx = np.unique(invid, return_index=True)
            n = idx.size
            if n >= unique.shape[0] or not np.all(np.isnan(unique[n:])):
                raise ValueError("`mat_buffer_size` too small")
            uids = indices[idx]
            uids = gridf_at.flatindex2index(uids[np.newaxis, :])
            uindices.append(uids)
            invindices.append(inv)
            indexmaps.append(_IdxMap(shift, gridf_at.index2flatindex))

        return self.replace(
            _cim=_CompressedIndexMap(
                base_kernel=None,
                kernels=None,
                uindices=uindices,
                indexmaps=indexmaps,
                invindices=invindices,
            )
        )

    def compress_matrices(self):
        """Recompute the kernel matrices while keeping the tables for speedy lookups
        fixed.

        This is useful for updating the covariance function of a kernel without
        recreating the tables in an expensive optimization.
        """
        assert self._cim is not None

        base_kernel = self.compute_matrices(jnp.array([-1]), None)
        kernels = tuple(
            self.compute_matrices(ii, ll) for ll, ii in enumerate(self._cim.uindices)
        )
        cim = self._cim._replace(base_kernel=base_kernel, kernels=kernels)
        return self.replace(_cim=cim)

    def compress(self, *args, **kwargs):
        """Optimize the kernel matrices for the given grid.

        This links kernel matrices by similarity and subsequently looks them up via
        tables. Compressing the kernel in this way has the potential to drastically
        reduce the memory requirements of applying it and make the inference of
        kernels faster.
        """
        return self.compress_indices(*args, **kwargs).compress_matrices()


def _default_window_size(grid_at_level, default=3) -> Tuple[int]:
    wsz = []
    for g in grid_at_level.raw_grids:
        if isinstance(g, HEALPixGridAtLevel):  # special-case HEALPix
            wsz += [9]
        elif isinstance(g, OpenGridAtLevel):
            wsz += list(g.padding * 2 + 1)
        else:
            wsz += [default] * g.ndim
    return tuple(wsz)


def refinement_matrices(cov, n_fsz: int):
    cov_ff = cov[-n_fsz:, -n_fsz:]
    cov_fc = cov[-n_fsz:, :-n_fsz]
    cov_cc = cov[:-n_fsz, :-n_fsz]

    olf = solve(cov_cc, cov_fc.T, matrix_eqn=True)
    return olf.T, sqrtm(cov_ff - cov_fc @ olf)


class ICRKernel(Kernel):
    """Full ICR implementation taking an arbitrary grid and a covariance function."""

    def __init__(self, grid, covariance, *, window_size=None, _cim=None):
        self._covariance_elem = covariance
        if window_size is None:
            window_size = tuple(
                _default_window_size(grid.at(lvl)) for lvl in range(grid.depth)
            )
        self._window_size = tuple(
            np.broadcast_to(window_size, (grid.depth, grid.at(0).ndim))
        )
        super().__init__(grid=grid, _cim=_cim)

    def replace(self, *, covariance=None, window_size=None, _cim=None):
        cim = self._cim if _cim is None else _cim
        if covariance is not None and cim is not None:
            cim = cim._replace(base_kernel=None, kernels=None)
        else:
            covariance = self._covariance_elem
        window_size = self.window_size if window_size is None else window_size
        return self.__class__(self.grid, covariance, window_size=window_size, _cim=cim)

    @property
    def covariance_outer(self):
        k = self._covariance_elem
        k = vmap(k, in_axes=(None, -1), out_axes=-1)
        k = vmap(k, in_axes=(-1, None), out_axes=-1)
        return k

    @property
    def window_size(self):
        return self._window_size

    def get_output_input_indices(self, index, level):
        if level is None:
            grid_at_lvl = self.grid.at(0)
            pixel_indices = np.mgrid[tuple(slice(0, sz) for sz in grid_at_lvl.shape)]
            assert pixel_indices.shape[0] == grid_at_lvl.ndim
            return (pixel_indices, 0), (
                (pixel_indices.reshape(grid_at_lvl.ndim, -1), 0),
            )
        g = self.grid.at(level)
        assert index.shape[0] == g.ndim
        gc = g.neighborhood(index, self.window_size[level]).reshape(index.shape + (-1,))
        gout = g.children(index)
        gf = gout.reshape(index.shape + (-1,))
        return (gout, level + 1), ((gc, level), (gf, level + 1))

    def compute_matrices(self, index, level):
        if level is None:
            _, ((ids, _),) = self.get_output_input_indices(index, None)
            gc = self.grid.at(0).index2coord(ids)
            cov = self.covariance_outer(gc, gc)
            assert cov.shape == (gc.shape[1],) * 2
            return (sqrtm(cov),)

        _, ((idc, _), (idf, _)) = self.get_output_input_indices(index, level)

        def get_mat(gc, gf):
            gc = self.grid.at(level).index2coord(gc)
            gf = self.grid.at(level + 1).index2coord(gf)
            assert gc.shape[0] == gf.shape[0]
            assert gc.ndim == gf.ndim == 2
            coord = jnp.concatenate((gc, gf), axis=-1)
            cov = self.covariance_outer(coord, coord)
            return refinement_matrices(cov, gf.shape[1])

        f = get_mat
        for _ in range(index.ndim - 1):
            f = vmap(f, in_axes=(1, 1))
        return f(idc, idf)
